#===============================================================================
# Copyright 2012 NetApp, Inc. All Rights Reserved,
# contribution by Jorge Mora <mora@netapp.com>
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#===============================================================================
"""
NFS utilities module

Provides a set of tools for testing NFS including methods for starting a packet
trace, stopping the packet trace and then open the packet trace for analysis.
It also provides a mechanism to enable NFS/RPC kernel debug and saving the
log messages for further analysis.

Furthermore, methods for finding specific NFSv4 operations within the packet
trace are also included.
"""
import os
from host import Host
from formatstr import *
import nfstest_config as c
from packet.nfs.nfs3_const import *
from packet.nfs.nfs4_const import *
from nfstest.utils import split_path

# Module constants
__author__    = "Jorge Mora (%s)" % c.NFSTEST_AUTHOR_EMAIL
__copyright__ = "Copyright (C) 2012 NetApp, Inc."
__license__   = "GPL v2"
__version__   = "2.7"

class NFSUtil(Host):
    """NFSUtil object

       NFSUtil() -> New NFSUtil object

       Usage:
           from nfstest.nfs_util import NFSUtil

           # Create object for local host
           x = NFSUtil()

           # Create client host object
           clientobj = x.create_host('192.168.0.11')

           # Use buffered matching on packets
           x.set_pktlist()

           # Get the next LOOKUP packets
           pktcall, pktreply = x.find_nfs_op(OP_LOOKUP)

           # Get OPEN information for the given file name
           fh, open_stid, deleg_stid = x.find_open(filename="file1")

           # Get address and port number from universal address string
           ipaddr, port = x.get_addr_port(addr)

           # Get packets and DS list for GETDEVICEINFO
           pktcall, pktreply, dslist = x.find_getdeviceinfo()

           # Get packets for EXCHANGE_ID
           pktcall, pktreply = x.find_exchange_id()

           # Get the NFS operation object from the given packet
           getfh = x.getop(x.pktreply, OP_GETFH)

           # Get the stateid which must be used by I/O operations
           stateid = x.get_stateid("file1")

           # Get the client id
           clientid = x.get_clientid()

           # Get the session id for the given clientid
           sessionid = x.get_sessionid(clientid=clientid)

           # Get the root file handle from PUTROOTFH for the given session id
           x.get_rootfh(sessionid=sessionid)

           # Get the file handle for the given path
           dirfh = x.get_pathfh("/vol1/data")

           # Display the state id in CRC16 format
           stidstr = x.stid_str(stateid)

           # Get the number of bytes available in the given directory
           freebytes = x.get_freebytes("/mnt/t")
    """
    def __init__(self, **kwargs):
        """Constructor

           Initialize object's private data.
        """
        # Arguments
        self.pktcall   = None
        self.pktreply  = None
        self.opencall  = None
        self.openreply = None

        # Initialize object variables
        self.clients = []
        self.clientobj = None
        self.nii_name = ''    # nii_name for the client
        self.nii_server = ''  # nii_name for the server
        self.device_info = {}
        self.dslist = []
        self.stateid = None
        self.rootfh  = None
        self.rootfsid = None
        self.rootfh_map = {} # Root fh map {key:sessionid, value:rootfh}
        self.sessionid_map = {} # Session id map {key:exchangeid, value:sessionid}
        self.sessionid = None # Session ID returned from CREATE_SESSION
        self.clientid = None # Client ID returned from EXCHANGE_ID

        # State id to string mapping
        self.stid_map = {}

        # Call base class constructor
        super(NFSUtil, self).__init__()

        # Initialize all test variables
        self.writeverf    = None
        self.test_seqid   = True
        self.test_stateid = True
        self.test_pattern = True
        self.test_niomiss = 0
        self.test_stripe  = True
        self.test_verf    = True
        self.need_commit  = False
        self.need_lcommit = False
        self.mdsd_lcommit = False
        self.max_iosize   = 0
        self.error_hash   = {}
        self.test_commit_full = True
        self.test_no_commit   = False
        self.test_commit_verf = True

    def __del__(self):
        """Destructor

           Gracefully stop the packet trace and un-reference all client
           objects
        """
        self.clientobj = None
        while self.clients:
            self.clients.pop()
        # Call base class destructor
        super(NFSUtil, self).__del__()

    def create_host(self, host, **kwargs):
        """Create client host object and set defaults."""
        self.clientobj = Host(
            host         = host,
            user         = kwargs.pop("user", ""),
            server       = kwargs.pop("server",       self.server),
            nfsversion   = kwargs.pop("nfsversion",   self.nfsversion),
            proto        = kwargs.pop("proto",        self.proto),
            port         = kwargs.pop("port",         self.port),
            sec          = kwargs.pop("sec",          self.sec),
            export       = kwargs.pop("export",       self.export),
            mtpoint      = kwargs.pop("mtpoint",      self.mtpoint),
            datadir      = kwargs.pop("datadir",      self.datadir),
            mtopts       = kwargs.pop("mtopts",       self.mtopts),
            nomount      = kwargs.pop("nomount",      self.nomount),
            tracename    = kwargs.pop("tracename",    self.tracename),
            trcdelay     = kwargs.pop("trcdelay",     self.trcdelay),
            tcpdump      = kwargs.pop("tcpdump",      self.tcpdump),
            tbsize       = kwargs.pop("tbsize",       self.tbsize),
            notrace      = kwargs.pop("notrace",      self.notrace),
            rpcdebug     = kwargs.pop("rpcdebug",     self.rpcdebug),
            nfsdebug     = kwargs.pop("nfsdebug",     self.nfsdebug),
            dbgname      = kwargs.pop("dbgname",      self.dbgname),
            messages     = kwargs.pop("messages",     self.messages),
            tmpdir       = kwargs.pop("tmpdir",       self.tmpdir),
            iptables     = kwargs.pop("iptables",     self.iptables),
            sudo         = kwargs.pop("sudo",         self.sudo),
        )

        self.clients.append(self.clientobj)
        return self.clientobj

    def set_pktlist(self, ops=None, cbs=None, procs=None, maxindex=None, pktdisp=False):
        """Set the current packet list for buffered matching in which the
           match method will only use this list instead of getting the next
           packet from the packet trace file. The default is to get all
           packets unless any of the arguments is given.

           NOTE: all READ reply data and all WRITE request data is discarded
           to avoid having memory issues.

           ops:
               List of NFSv4 operations to include in the packet list
           cbs:
               List of NFSv4 callback operations to include in the packet list
           procs:
               List of NFSv3 procedures to include in the packet list
           maxindex:
               Include packets up to but not including the packet indexed
               by this argument [default: None]
               A value of None means there is no limit
           pktdisp:
               Display all cached packets [default: False]
        """
        pktlist = []
        # Default behavior when no list is given
        defexpr = ops is None and cbs is None and procs is None
        # Boolean expressions for each of the lists
        ops_expr   = not defexpr and ops   is not None
        cbs_expr   = not defexpr and cbs   is not None
        procs_expr = not defexpr and procs is not None
        for pkt in self.pktt:
            # Get list of NFS packets
            if pkt == "nfs":
                if maxindex is not None and pkt.record.index >= maxindex:
                    break

                rpc = pkt.rpc
                if rpc.procedure == 0:
                    # NULL procedure
                    if not defexpr and (not procs_expr or 0 not in procs):
                        continue
                elif (rpc.version == 4 and not pkt.nfs.callback) or \
                     (rpc.version == 1 and pkt.nfs.callback):
                    # NFSv4 COMPOUND and callback
                    incl_pkt = False
                    for item in pkt.nfs.array:
                        op = item.op
                        # Discard data from read and write packets so memory
                        # is not an issue. Do this before selecting operations
                        # in case a READ or WRITE packet is selected by any
                        # of the other operations in the array
                        if op == OP_READ and rpc.type == 1:
                            item.opread.resok.data = ""
                        elif op == OP_WRITE and rpc.type == 0:
                            item.opwrite.data = ""
                        if not defexpr:
                            # If any of the lists is given, make sure to
                            # include only operations in the given lists
                            if pkt.nfs.callback:
                                if not cbs_expr or op not in cbs:
                                    continue
                            else:
                                if not ops_expr or op not in ops:
                                    continue
                        incl_pkt = True
                    if not incl_pkt:
                        continue
                elif rpc.version == 3:
                    # NFSv3 procedures
                    procedure = pkt.nfs.procedure
                    # If the procs list is given, make sure to include only
                    # procedures given in the list
                    if not defexpr and (not procs_expr or procedure not in procs):
                        continue
                    # Discard data from read and write packets
                    # so memory is not an issue
                    if procedure == NFSPROC3_READ and rpc.type == 1:
                        pkt.nfs.opread.resok.data = ""
                    elif procedure == NFSPROC3_WRITE and rpc.type == 0:
                        pkt.nfs.opwrite.data = ""
                pktlist.append(pkt)
                if pktdisp:
                    self.test_info(str(pkt))
        self.pktt.set_pktlist(pktlist)

    def find_nfs_op(self, op, **kwargs):
        """Find the call and its corresponding reply for the specified NFSv4
           operation going to the server specified by the ipaddr and port.
           The reply must also match the given status. Also the following
           object attributes are defined: pktcall referencing the packet call
           while pktreply referencing the packet reply.

           op:
               NFS operation to find
           ipaddr:
               Destination IP address [default: self.server_ipaddr]
               A value of None matches any IP address
           port:
               Destination port [default: self.port]
               A value of None matches any destination port
           match:
               Match string to include [default: '']
           status:
               Match the status of the operation [default: 0]
               A value of None matches any status.
           src_ipaddr:
               Source IP address [default: None]
               A value of None matches any IP address
           maxindex:
               The match fails if packet index hits this limit [default: None]
               A value of None means there is no limit
           call_only:
               Find the call only [default: False]

           Return a tuple: (pktcall, pktreply).
        """
        ipaddr       = kwargs.get("ipaddr",       self.server_ipaddr)
        port         = kwargs.get("port",         self.port)
        match        = kwargs.get("match",        "")
        status       = kwargs.get("status",       0)
        src_ipaddr   = kwargs.get("src_ipaddr",   None)
        maxindex     = kwargs.get("maxindex",     None)
        call_only    = kwargs.get("call_only",    False)

        mstatus = "" if status is None else "NFS.status == %d and " % status
        src = "IP.src == '%s' and " % src_ipaddr if src_ipaddr != None else ''
        dst = "IP.dst == '%s' and " % ipaddr if ipaddr is not None else ""
        if len(match):
            match += " and "
        if port != None:
            dst += "TCP.dst_port == %d and " % port
        pktcall  = None
        pktreply = None
        while True:
            # Find request
            pktcall = self.pktt.match(src + dst + match + "NFS.argop == %d" % op, maxindex=maxindex)
            if pktcall and not call_only:
                # Find reply
                xid = pktcall.rpc.xid
                pktreply = self.pktt.match("RPC.xid == %d and %s NFS.resop == %d" % (xid, mstatus, op), maxindex=maxindex)
                if pktreply:
                    break
            else:
                break
        self.pktcall  = pktcall
        self.pktreply = pktreply
        return (pktcall, pktreply)

    def find_open(self, **kwargs):
        """Find the call and its corresponding reply for the NFSv4 OPEN of the
           given file going to the server specified by the ipaddr and port.
           The following object attributes are defined: opencall and pktcall
           both referencing the packet call while openreply and pktreply both
           referencing the packet reply.

           filename:
               Find open call and reply for this file [default: None]
           claimfh:
               Find open call and reply for this file handle using CLAIM_FH
               [default: None]
           ipaddr:
               Destination IP address [default: self.server_ipaddr]
           port:
               Destination port [default: self.port]
           deleg_type:
               Expected delegation type on reply [default: None]
           deleg_stateid:
               Delegation stateid expected on call in delegate_cur_info [default: None]
           fh:
               Find open call and reply for this file handle when using
               deleg_stateid or as the directory FH when deleg_stateid
               is not set [default: None]
           src_ipaddr:
               Source IP address [default: any IP address]
           maxindex:
               The match fails if packet index hits this limit [default: no limit]
           anyclaim:
               Find open for either regular open or using delegate_cur_info [default: False]

           Must specify either filename, claimfh or both.
           Return a tuple: (filehandle, open_stateid, deleg_stateid).
        """
        filename      = kwargs.pop('filename', None)
        claimfh       = kwargs.pop('claimfh', None)
        fh            = kwargs.pop('fh', None)
        ipaddr        = kwargs.pop('ipaddr', self.server_ipaddr)
        port          = kwargs.pop('port', self.port)
        deleg_type    = kwargs.pop('deleg_type', None)
        deleg_stateid = kwargs.pop('deleg_stateid', None)
        src_ipaddr    = kwargs.pop('src_ipaddr', None)
        maxindex      = kwargs.pop('maxindex', None)
        anyclaim      = kwargs.pop('anyclaim', False)
        self.pktcall  = None
        self.pktreply = None
        self.opencall  = None
        self.openreply = None

        src = "IP.src == '%s' and " % src_ipaddr if src_ipaddr is not None else ''
        dst = self.pktt.ip_tcp_dst_expr(ipaddr, port)

        file_str = ""
        deleg_str = ""
        claimfh_str = ""
        str_list = []
        if filename is not None:
            file_str = "NFS.claim.name == '%s'" % filename
            str_list.append(file_str)
        if claimfh is not None:
            claimfh_str = "(NFS.fh == '%s' and NFS.claim.claim == %d)" % (self.pktt.escape(claimfh), CLAIM_FH)
            str_list.append(claimfh_str)
        if deleg_stateid is not None:
            deleg_str  = "(NFS.claim.claim == %d" % CLAIM_DELEGATE_CUR
            deleg_str += " and NFS.claim.deleg_info.name == '%s'" % filename
            deleg_str += " and NFS.claim.deleg_info.stateid == '%s')" % self.pktt.escape(deleg_stateid)
            if fh is not None:
                deleg_str += " or (NFS.claim.claim == %d" % CLAIM_DELEG_CUR_FH
                deleg_str += " and NFS.fh == '%s' and NFS.claim.stateid == '%s')" % (self.pktt.escape(fh), self.pktt.escape(deleg_stateid))
            str_list.append("(" + deleg_str + ")")
        if claimfh is None and deleg_stateid is None and fh is not None:
            dirfh_str = "NFS.fh == '%s'" % self.pktt.escape(fh)
            file_str = dirfh_str + " and " + file_str
            str_list.append(dirfh_str)

        if anyclaim:
            file_str = " or ".join(str_list)
        elif claimfh is not None:
            file_str = claimfh_str
        elif deleg_stateid is not None:
            file_str = deleg_str
        elif len(file_str) == 0:
            raise Exception("Must specify either filename or claimfh")

        while True:
            pktcall = self.pktt.match(src + dst + " and NFS.argop == %d and %s" % (OP_OPEN, file_str), maxindex=maxindex)
            if not pktcall:
                return (None, None, None)
            xid = pktcall.rpc.xid
            open_str = "RPC.xid == %d and NFS.status == 0 and NFS.resop == %d" % (xid, OP_OPEN)
            if deleg_type is not None:
                open_str += " and NFS.delegation.deleg_type == %d" % deleg_type

            # Find OPEN reply to get filehandle of file
            pktreply = self.pktt.match(open_str, maxindex=maxindex)
            if not pktreply:
                continue

            if claimfh is None:
                # GETFH should be the operation following the OPEN,
                # but look for it just in case it is not
                idx = pktreply.NFSidx + 1
                resarray = pktreply.nfs.array
                while (idx < len(resarray) and resarray[idx].resop != OP_GETFH):
                    idx += 1
                if idx >= len(resarray):
                    # Could not find GETFH
                    if fh is None:
                        return (None, None, None)
                    else:
                        filehandle = fh
                else:
                    filehandle = pktreply.nfs.array[idx].fh
            else:
                # No need to find GETFH, the filehandle is already known
                filehandle = claimfh

            open_stateid = pktreply.NFSop.stateid.other
            if pktreply.NFSop.delegation.deleg_type in [OPEN_DELEGATE_READ, OPEN_DELEGATE_WRITE]:
                deleg_stateid = pktreply.NFSop.delegation.stateid.other
            else:
                deleg_stateid = None

            self.pktcall  = pktcall
            self.pktreply = pktreply
            self.opencall  = pktcall
            self.openreply = pktreply
            return (filehandle, open_stateid, deleg_stateid)

    def find_layoutget(self, filehandle):
        """Find the call and its corresponding reply for the NFSv4 LAYOUTGET
           of the given file handle going to the server specified by the
           ipaddr for self.server and port given by self.port.

           Return a tuple: (layoutget, layoutget_res, loc_body).
        """
        dst = self.pktt.ip_tcp_dst_expr(self.server_ipaddr, self.port)

        # Find LAYOUTGET request
        pkt = self.pktt.match(dst + " and NFS.fh == '%s' and NFS.argop == %d" % (self.pktt.escape(filehandle), OP_LAYOUTGET))
        if not pkt:
            return (None, None, None)
        xid = pkt.rpc.xid
        layoutget = pkt.NFSop

        # Find LAYOUTGET reply
        pkt = self.pktt.match("RPC.xid == %d and NFS.resop == %d" % (xid, OP_LAYOUTGET))
        if pkt is None:
            return (layoutget, None, None)
        layoutget_res = pkt.NFSop
        if layoutget_res.status:
            return (layoutget, layoutget_res, None)
        # XXX Using first layout segment only
        layout = layoutget_res.layout[0]

        # Get layout content
        loc_body = layout.content.body
        if layoutget.type == LAYOUT4_NFSV4_1_FILES:
            nfl_util = loc_body.nfl_util

            # Decode loc_body
            loc_body = {
                'type':               layoutget.type,
                'dense':              (nfl_util & NFL4_UFLG_DENSE > 0),
                'commit_mds':         (nfl_util & NFL4_UFLG_COMMIT_THRU_MDS > 0),
                'stripe_size':        nfl_util & NFL4_UFLG_STRIPE_UNIT_SIZE_MASK,
                'first_stripe_index': loc_body.first_stripe_index,
                'offset':             loc_body.pattern_offset,
                'filehandles':        loc_body.fh_list,
                'deviceid':           loc_body.deviceid,
                'stateid':            layoutget_res.stateid.other,
                'iomode':             layout.iomode,
            }
        else:
            loc_body = {
                'type':               layoutget.type,
                'stateid':            layoutget_res.stateid.other,
                'iomode':             layout.iomode,
            }

        return (layoutget, layoutget_res, loc_body)

    def get_addr_port(self, addr):
        """Get address and port number from universal address string"""
        addr_list = addr.split('.')
        if len(addr_list) == 6:
            # IPv4 address
            ipaddr = '.'.join(addr_list[:4])
        else:
            # IPv6 address
            ipaddr = addr_list[0]
        port = (int(addr_list[-2])<<8) + int(addr_list[-1])
        return ipaddr, port

    def find_getdeviceinfo(self, deviceid=None):
        """Find the call and its corresponding reply for the NFSv4 GETDEVICEINFO
           going to the server specified by the ipaddr for self.server and port
           given by self.port.

           deviceid:
               Look for an specific deviceid [default: any deviceid]

           Return a tuple: (pktcall, pktreply, dslist).
        """
        dslist = []
        # Find GETDEVICEINFO request and reply
        match = "NFS.deviceid == '%s'" % self.pktt.escape(deviceid) if deviceid is not None else ''
        (pktcall, pktreply) = self.find_nfs_op(OP_GETDEVICEINFO, match=match, status=None)
        if pktreply and pktreply.nfs.status == 0:
            self.gdir_device = pktreply.NFSop.device_addr
            if self.gdir_device.type == LAYOUT4_NFSV4_1_FILES:
                da_addr_body = self.gdir_device.body
                self.stripe_indices = da_addr_body.stripe_indices
                multipath_ds_list = da_addr_body.multipath_ds_list

                for ds_list in multipath_ds_list:
                    dslist.append([])
                    for item in ds_list:
                        # Get ip address and port for DS
                        ipaddr, port = self.get_addr_port(item.addr)
                        dslist[-1].append({'ipaddr': ipaddr, 'port': port})
            # Save device info for future reference
            self.device_info[pktcall.NFSop.deviceid] = {
                'call':  pktcall,
                'reply': pktreply,
            }
        if len(dslist) > 0:
            self.dslist = dslist
        return (pktcall, pktreply, dslist)

    def find_exchange_id(self, **kwargs):
        """Find the call and its corresponding reply for the NFSv4 EXCHANGE_ID
           going to the server specified by the ipaddr and port.

           ipaddr:
               Destination IP address [default: self.server_ipaddr]
           port:
               Destination port [default: self.port]

           Store the callback IP/TCP expression in object attribute cb_dst

           Return a tuple: (pktcall, pktreply).
        """
        # Find EXCHANGE_ID request and reply
        (pktcall, pktreply) = self.find_nfs_op(OP_EXCHANGE_ID, **kwargs)
        self.src_ipaddr = pktcall.ip.src
        self.src_port   = pktcall.tcp.src_port
        self.cb_dst     = self.pktt.ip_tcp_dst_expr(self.src_ipaddr, self.src_port)

        if pktcall is not None and pktcall.NFSop.client_impl_id is not None:
            self.nii_name = pktcall.NFSop.client_impl_id.name
        if pktreply is not None and pktreply.NFSop.server_impl_id is not None:
            self.nii_server = pktreply.NFSop.server_impl_id.name

        return (pktcall, pktreply)

    def find_layoutrecall(self, status=0):
        """Find NFSv4 CB_LAYOUTRECALL call and return its reply.
           The reply must also match the given status.
        """
        # Find CB_LAYOUTRECALL request
        pktcall = self.pktt.match(self.cb_dst + " and NFS.argop == %d" % OP_CB_LAYOUTRECALL)
        if pktcall:
            # Find reply
            xid = pktcall.rpc.xid
            pktreply = self.pktt.match("RPC.xid == %d and NFS.resop == %d and NFS.status == %d" % (xid, OP_CB_LAYOUTRECALL, status))
        else:
            self.test(False, "CB_LAYOUTRECALL was not found")
            return
        return pktreply

    def get_abs_offset(self, offset, ds_index=None):
        """Get real file offset given by the (read/write) offset on the given
           data server index, taking into account the type of layout
           (dense/sparse), the stripe_size, first stripe index and the number
           of filehandles. The layout information is taken from object
           attribute layout.
        """
        if ds_index is None:
            return offset
        nfhs = len(self.dslist)
        stripe_size = self.layout['stripe_size']
        first_stripe_index = self.layout['first_stripe_index']
        ds_index -= first_stripe_index;
        if ds_index < 0:
            ds_index += nfhs
        # Get real file offset given by the read/write offset to the given DS index
        if self.layout['dense']:
            # Dense layout
            n = int(offset / stripe_size)
            r = offset % stripe_size
            file_offset = (n*nfhs + ds_index)*stripe_size + r
        else:
            # Sparse layout
            file_offset = offset
        return file_offset

    def get_filehandle(self, ds_index):
        """Return filehandle from the layout list of filehandles."""
        if len(self.layout['filehandles']) > 1:
            filehandle = self.layout['filehandles'][ds_index]
        else:
            filehandle = self.layout['filehandles'][0]
        return filehandle

    def verify_stripe(self, offset, size, ds_index):
        """Verify if read/write is sent to the correct data server according
           to stripe size, first stripe index and the number of filehandles.
           The layout information is taken from object attribute layout.

           offset:
               Real file offset
           size:
               I/O size
           ds_index:
               Data server index

           Return True if stripe is correctly verified, False otherwise.
        """
        nfhs = len(self.dslist)
        if self.layout is None or ds_index is None:
            return False
        stripe_size = self.layout['stripe_size']
        first_stripe_index = self.layout['first_stripe_index']
        n = int(offset / stripe_size)
        m = int((offset + size - 1) / stripe_size)
        idx = n % nfhs
        ds_index -= first_stripe_index;
        if ds_index < 0:
            ds_index += nfhs
        return n == m and idx == ds_index

    def getop(self, pkt, op):
        """Get the NFS operation object from the given packet"""
        if pkt:
            # Start looking for the operation after NFSidx if it exists
            if hasattr(pkt, "NFSidx"):
                idx = pkt.NFSidx + 1
            else:
                idx = 0
            array = pkt.nfs.array
            while (idx < len(array) and array[idx].op != op):
                idx += 1
            if idx < len(array):
                # Return the operation object
                return pkt.nfs.array[idx]
        return

    def verify_pnfs_supported(self, filehandle, server_type, path=None):
        """Verify pNFS is supported in the given server path.
           Finds the GETATTR asking for FATTR4_SUPPORTED_ATTRS(bit 0 and its
           reply to verify FATTR4_FS_LAYOUT_TYPES is supported for the path.
           Then it finds the GETATTR asking for FATTR4_FS_LAYOUT_TYPES(bit 62)
           to verify LAYOUT4_NFSV4_1_FILES is returned in fs_layout_types.
        """
        if path:
            pmsg = " for %s" % path
        else:
            pmsg = ""
        fhstr = self.pktt.escape(filehandle)
        # Find packet having a GETATTR asking for FATTR4_SUPPORTED_ATTRS(bit 0)
        attrmatch = "NFS.fh == '%s' and NFS.request & %s != 0" % (fhstr, hex(1 << FATTR4_SUPPORTED_ATTRS))
        pktcall, pktreply = self.find_nfs_op(OP_GETATTR, match=attrmatch)
        self.test(pktcall, "GETATTR should be sent to %s asking for FATTR4_SUPPORTED_ATTRS%s" % (server_type, pmsg))
        if pktreply:
            supported_attrs = pktreply.NFSop.attributes[FATTR4_SUPPORTED_ATTRS]
            fslt_supported = supported_attrs & (1<<FATTR4_FS_LAYOUT_TYPES) != 0
            self.test(fslt_supported, "NFS server should support pNFS layout types (FATTR4_FS_LAYOUT_TYPES)%s" % pmsg)
        elif pktcall:
            self.test(False, "GETATTR reply was not found")

        # Find packet having a GETATTR asking for FATTR4_FS_LAYOUT_TYPES(bit 62)
        attrmatch = "NFS.fh == '%s' and NFS.request & %s != 0" % (fhstr, hex(1 << FATTR4_FS_LAYOUT_TYPES))
        pktcall, pktreply = self.find_nfs_op(OP_GETATTR, match=attrmatch)
        self.test(pktcall, "GETATTR should be sent to %s asking for FATTR4_FS_LAYOUT_TYPES%s" % (server_type, pmsg))
        if pktreply:
            # Get list of fs layout types supported by the server
            fs_layout_types = pktreply.NFSop.attributes[FATTR4_FS_LAYOUT_TYPES]
            self.test(LAYOUT4_NFSV4_1_FILES in fs_layout_types, "NFS server should return LAYOUT4_NFSV4_1_FILES in fs_layout_types%s" % pmsg)
        elif pktcall:
            self.test(False, "GETATTR reply was not found")

    def verify_create_session(self, ipaddr, port, ds=False, nocreate=False, ds_index=None, exchid_status=0, cs_status=0):
        """Verify initial connection to the metadata server(MDS)/data server(DS).
           Verify if EXCHANGE_ID, CREATE_SESSION, RECLAIM_COMPLETE,
           GETATTR asking for FATTR4_LEASE_TIME, and GETATTR asking for
           FATTR4_FS_LAYOUT_TYPES are all sent or not to the server.

           ipaddr:
               Destination IP address of MDS or DS
           port:
               Destination port number of MDS or DS
           ds:
               True if ipaddr/port defines a DS, otherwise MDS [default: False]
           nocreate:
               True if expecting the client NOT to send EXCHANGE_ID,
               CREATE_SESSION, and RECLAIM_COMPLETE. Otherwise, verify all
               these operations are sent by the client [default: False]
           ds_index:
               DS index used for displaying purposes only [default: None]
           exchid_status:
               Expected status for EXCHANGE_ID [default: 0]
           cs_status:
               Expected status for CREATE_SESSION [default: 0]

           Return the sessionid and it is also stored in the object
           attribute sessionid.
        """
        self.sessionid = None
        if ds:
            pnfs_use_flag = EXCHGID4_FLAG_USE_PNFS_DS
            server_type = "DS"
            if ds_index is not None:
                server_type += "(%d)" % ds_index
        else:
            pnfs_use_flag = EXCHGID4_FLAG_USE_PNFS_MDS
            server_type = "MDS"

        dsmds = ""
        if ds_index != None and ipaddr == self.server_ipaddr and port == self.port:
            # DS == MDS, client does not connect to DS, it has a connection already
            nocreate = True
            dsmds = " since DS == MDS"

        if not ds:
            save_index = self.pktt.get_index()
            # Find PUTROOTFH having a GETFH operation
            getfhmatch = "NFS.argop == %d" % OP_GETFH
            pktcall, pktreply = self.find_nfs_op(OP_PUTROOTFH, ipaddr=ipaddr, port=port, match=getfhmatch)
            self.rootfh = getattr(self.getop(pktreply, OP_GETFH), "fh", None)
            attributes  = getattr(self.getop(pktreply, OP_GETATTR), "attributes", None)
            if attributes:
                self.rootfsid = attributes.get(FATTR4_FSID)
            self.pktt.rewind(save_index)

        # Find EXCHANGE_ID request and reply
        (pktcall, pktreply) = self.find_nfs_op(OP_EXCHANGE_ID, ipaddr=ipaddr, port=port, status=exchid_status)
        if nocreate:
            self.test(not pktcall, "EXCHANGE_ID should not be sent to %s%s" % (server_type, dsmds))
        else:
            self.test(pktcall, "EXCHANGE_ID should be sent to %s" % server_type)
            if pktreply:
                if exchid_status:
                    self.test(pktreply.NFSop.status == exchid_status, "EXCHANGE_ID reply should return %s(%d)" % (nfsstat4[exchid_status], exchid_status))
                    return
                else:
                    eir_flags = pktreply.NFSop.flags
                    if pktreply.NFSop.server_impl_id is not None:
                        self.nii_name = pktreply.NFSop.server_impl_id.name
                    self.test(eir_flags & pnfs_use_flag != 0, "EXCHGID4_FLAG_USE_PNFS_%s should be set" % server_type, terminate=True)
                    if not ds:
                        # Check for invalid combination of eir flags
                        self.test(eir_flags & EXCHGID4_FLAG_USE_NON_PNFS == 0, "EXCHGID4_FLAG_USE_NON_PNFS should not be set")
            else:
                self.test(False, "EXCHANGE_ID reply was not found")

        # Find CREATE_SESSION request
        (pktcall, pktreply) = self.find_nfs_op(OP_CREATE_SESSION, ipaddr=ipaddr, port=port, status=cs_status)
        if nocreate:
            self.test(not pktcall, "CREATE_SESSION should not be sent to %s%s" % (server_type, dsmds))
        else:
            self.test(pktcall, "CREATE_SESSION should be sent to %s" % server_type)
            if pktreply:
                if cs_status:
                    self.test(pktreply.NFSop.status == cs_status, "CREATE_SESSION reply should return %s(%d)" % (nfsstat4[cs_status], cs_status))
                    return
                else:
                    # Save the session id
                    self.sessionid = pktreply.NFSop.sessionid
                    # Save the max response size
                    self.ca_maxrespsz = pktreply.NFSop.fore_chan_attrs.maxresponsesize
                    self.dprint('DBG2', "CREATE_SESSION sessionid: %s" % self.sessionid)
                    self.dprint('DBG2', "CREATE_SESSION ca_maxrespsz: %s" % self.ca_maxrespsz)

                    fmsg = None
                    test_seq = True
                    slotid_map = {}
                    save_index = self.pktt.get_index()
                    while self.find_nfs_op(OP_SEQUENCE, ipaddr=ipaddr, port=port, call_only=True):
                        if self.pktcall is None:
                            break
                        slotid = self.pktcall.NFSop.slotid
                        seqid  = self.pktcall.NFSop.sequenceid
                        if slotid_map.get(slotid) is None:
                            # First occurrence of slot id
                            slotid_map[slotid] = seqid
                            if seqid != 1:
                                fmsg = ", slot id %d starts with sequence id %d" % (slotid, seqid)
                                test_seq = False
                                break
                    if len(slotid_map) > 0:
                        self.test(test_seq, "SEQUENCE request should start with a sequence id of 1", failmsg=fmsg)
                    else:
                        self.test(False, "SEQUENCE request was not found")
                    self.pktt.rewind(save_index)
            elif pktcall:
                self.test(False, "CREATE_SESSION reply was not found")

        # Find RECLAIM_COMPLETE request
        (pktcall, pktreply) = self.find_nfs_op(OP_RECLAIM_COMPLETE, ipaddr=ipaddr, port=port, status=None)
        if nocreate:
            self.test(not pktcall, "RECLAIM_COMPLETE should not be sent to %s%s" % (server_type, dsmds))
        else:
            self.test(pktcall, "RECLAIM_COMPLETE should be sent to %s" % server_type)
        if pktcall:
            # Make sure to start the next packet search right after the
            # RECLAIM_COMPLETE call
            self.pktt.rewind(pktcall.record.index)

        if not ds:
            # Find packet having a GETATTR asking for FATTR4_LEASE_TIME(bit 10)
            attrmatch = "NFS.request & %s != 0" % hex(1 << FATTR4_LEASE_TIME)
            (pktcall, pktreply) = self.find_nfs_op(OP_GETATTR, match=attrmatch)
            self.test(pktcall, "GETATTR should be sent to %s asking for FATTR4_LEASE_TIME" % server_type)
            if pktreply:
                lease_time = pktreply.NFSop.attributes[FATTR4_LEASE_TIME]
                self.test(lease_time > 0, "NFS server should return lease time(%d) > 0" % lease_time)
            elif pktcall:
                self.test(False, "GETATTR reply was not found")

            self.verify_pnfs_supported(self.rootfh, server_type)
            save_index = self.pktt.get_index()

            # Find if pNFS is supported for the mounted path including datadir
            path_list = []
            if len(self.datadir):
                path = os.path.join(self.export, self.datadir)
            else:
                path = self.export
            while True:
                plist = os.path.split(path)
                if plist[1] == "":
                    break
                path_list.insert(0, plist[1])
                path = plist[0]
            fullpath = "/"
            fsid_list = []
            if self.rootfsid is not None:
                fsid_list.append(self.rootfsid)
            for path in path_list:
                # Find the LOOKUP
                fullpath = os.path.join(fullpath, path)
                match = "NFS.name == '%s'" % path
                pktcall, pktreply = self.find_nfs_op(OP_LOOKUP, match=match)
                if pktreply:
                    getfh_obj   = self.getop(pktreply, OP_GETFH)
                    getattr_obj = self.getop(pktreply, OP_GETATTR)
                    if getfh_obj is None or getattr_obj is None:
                        # Could not find GETFH or GETATTR
                        continue
                    filehandle = getfh_obj.fh
                    attributes = getattr_obj.attributes
                    fsid = attributes.get(FATTR4_FSID)
                    for xfsid in fsid_list:
                        if fsid.major == xfsid.major and fsid.minor == xfsid.minor:
                            # This fsid has already been verified, so skip it
                            fsid = None
                            break
                    if fsid is not None:
                        # Save the fsid so it won't be verified again
                        fsid_list.append(fsid)
                        # Verify this path supports pNFS
                        self.verify_pnfs_supported(filehandle, server_type, path=fullpath)
            self.pktt.rewind(save_index)
        return self.sessionid

    def verify_layoutget(self, filehandle, iomode, riomode=None, status=0, offset=None, length=None):
        """Verify the client sends a LAYOUTGET for the given file handle.

           filehandle:
               Find LAYOUTGET for this file handle
           iomode:
               Expected I/O mode for LAYOUTGET call
           riomode:
               Expected I/O mode for LAYOUTGET reply if specified, else verify
               reply I/O mode is equal to call I/O mode if iomode == 2.
               If iomode == 1, the reply I/O mode could be equal to 1 or 2
           status:
               Expected status for LAYOUTGET reply [default: 0]
           offset:
               Expected layout range for LAYOUTGET reply [default: None]
           length:
               Expected layout range for LAYOUTGET reply [default: None]

           If both offset and length are not given, verify LAYOUTGET reply
           should be a full layout [0, NFS4_UINT64_MAX]. If only one is
           provided the following defaults are used: offset = 0,
           length = NFS4_UINT64_MAX.

           Layout information is stored in the object attribute layout.

           Return a tuple: (layoutget, layoutget_res, loc_body).
        """
        # Find LAYOUTGET for given filehandle
        layoutget, layoutget_res, loc_body = self.find_layoutget(filehandle)
        if layoutget is None:
            self.layout = None
            return (None, None, None)

        # Test layout type
        self.test(layoutget.type == LAYOUT4_NFSV4_1_FILES, "LAYOUTGET layout type should be LAYOUT4_NFSV4_1_FILES")

        # Test iomode
        self.test(layoutget.iomode == iomode, "LAYOUTGET iomode should be %s" % self.iomode_str(iomode))

        # Test for full file layout
        self.test(layoutget.offset == 0 and layoutget.length == NFS4_UINT64_MAX, "LAYOUTGET should ask for full file layout")

        if layoutget_res is None:
            self.test(False, "LAYOUTGET reply should be returned")
            return (layoutget, None, None)

        if status:
            self.test(layoutget_res.status == status, "LAYOUTGET reply should return error %s(%d)" % (nfsstat4[status], status))
            return (layoutget, layoutget_res, None)
        elif layoutget_res.status:
            self.test(False, "LAYOUTGET reply returned %s(%d)" % (nfsstat4[layoutget_res.status], layoutget_res.status))
            return (layoutget, layoutget_res, None)

        # Get layout from reply
        layout = layoutget_res.layout[0]

        # Test LAYOUTGET reply for correct layout type
        self.test(layout.content.type == LAYOUT4_NFSV4_1_FILES, "LAYOUTGET reply layout type should be LAYOUT4_NFSV4_1_FILES")

        # Test LAYOUTGET reply for correct iomode
        if riomode is not None:
            self.test(layout.iomode == riomode, "LAYOUTGET reply iomode is %s when asking for a %s layout" % (self.iomode_str(riomode), self.iomode_str(iomode)))
        elif iomode == LAYOUTIOMODE4_READ and layoutget.iomode in [LAYOUTIOMODE4_READ, LAYOUTIOMODE4_RW]:
            self.test(True, "LAYOUTGET reply iomode is %s when asking for a LAYOUTIOMODE4_READ layout" % self.iomode_str(layoutget.iomode))
        else:
            self.test(layout.iomode == iomode, "LAYOUTGET reply iomode should be %s" % self.iomode_str(iomode))

        if offset is None and length is None:
            # Test LAYOUTGET reply for full file layout
            self.test(layout.offset == 0 and layout.length == NFS4_UINT64_MAX, "LAYOUTGET reply should be full file layout")
        else:
            # Test LAYOUTGET reply for correct layout range
            if offset is None:
                offset = 0
            if length is None:
                length = NFS4_UINT64_MAX
            self.test(layout.offset == offset and layout.length == length, "LAYOUTGET reply should be: (offset=%d, length=%d)" % (offset, length))

        # Return layout
        self.layout = loc_body
        self.layout['return_on_close'] = layoutget_res.return_on_close
        return (layoutget, layoutget_res, loc_body)

    def verify_io(self, iomode, stateid, ipaddr=None, port=None, src_ipaddr=None, filehandle=None, ds_index=None, init=False, maxindex=None, pattern=None):
        """Verify I/O is sent to the server specified by the ipaddr and port.

           iomode:
               Verify reads (iomode == 1) or writes (iomode == 2)
           stateid:
               Expected stateid to use in all I/O requests
           ipaddr:
               Destination IP address of MDS or DS
               [default: do not match destination]
           port:
               Destination port number of MDS or DS
               [default: do not match destination port]
           src_ipaddr:
               Source IP address of request
               [default: do not match source]
           filehandle:
               Find I/O for this file handle. This option is used when
               verifying I/O sent to the MDS
               [default: use filehandle given by ds_index]
           ds_index:
               Data server index. This option is used when verifying I/O sent
               to the DS -- filehandle is taken from x.layout for this index
               [default: None]
           init:
               Initialized test variables [default: False]
           maxindex:
               The match fails if packet index hits this limit [default: no limit]
           pattern:
               Data pattern to compare [default: default data pattern]

           Return the number of I/O operations sent to the server.
        """
        if filehandle is None:
            filehandle = self.get_filehandle(ds_index)
        src = "IP.src == '%s' and " % src_ipaddr if src_ipaddr != None else ''
        dst = ''
        if ipaddr != None:
            dst = "IP.dst == '%s' and " % ipaddr
            if port != None:
                dst += "TCP.dst_port == %d and " % port
        fh = "NFS.fh == '%s'" % self.pktt.escape(filehandle)
        save_index = self.pktt.get_index()
        xids = []
        offsets = {}
        good_pattern = 0
        bad_pattern = 0
        self.test_offsets = []  # Save the offsets sent to the server on I/O
        self.test_counts = []   # Save the counts received from the server
        xid_counts = {}         # Map counts on I/O calls
        if init:
            self.test_seqid   = True
            self.test_stateid = True
            self.test_pattern = True
            self.test_niomiss = 0
            self.test_stripe  = True
            self.test_verf    = True
            self.need_commit  = False
            self.need_lcommit = False
            self.mdsd_lcommit = False
            self.stateid      = None
            self.max_iosize   = 0
            self.error_hash   = {}

        # Get I/O type: iomode == 1 (READ), else (WRITE)
        io_op = OP_READ if iomode == LAYOUTIOMODE4_READ else OP_WRITE

        # Find all I/O requests for MDS or current DS
        while True:
            # Find I/O request
            pkt = self.pktt.match(src + dst + fh + " and NFS.argop == %d" % io_op, maxindex=maxindex)
            if not pkt:
                break
            xids.append(pkt.rpc.xid)
            nfsop = pkt.NFSop
            self.test_offsets.append(nfsop.offset)
            xid_counts[pkt.rpc.xid] = nfsop.count
            if iomode == LAYOUTIOMODE4_READ:
                offsets[pkt.rpc.xid] = nfsop.offset

            if nfsop.stateid.seqid != 0:
                self.test_seqid = False
            if nfsop.stateid != stateid:
                self.test_stateid = False
            self.stateid = nfsop.stateid.other

            # Get real file offset
            file_offset = self.get_abs_offset(nfsop.offset, ds_index)

            size = nfsop.count
            if iomode != LAYOUTIOMODE4_READ:
                data = self.data_pattern(file_offset, len(nfsop.data), pattern=pattern)
                if data != nfsop.data:
                    bad_pattern += 1
                else:
                    good_pattern += 1
            if self.max_iosize < size:
                self.max_iosize = size

            # Check if I/O is sent to the MDS or correct DS according to stripe size
            if ds_index is not None and not self.verify_stripe(file_offset, size, ds_index):
                self.test_stripe = False

        # Rewind trace file to saved packet index
        self.pktt.rewind(save_index)

        if iomode == LAYOUTIOMODE4_RW:
            self.dprint('DBG7', "WRITE bad/good pattern %d/%d" % (bad_pattern, good_pattern))
            if good_pattern == 0 or float(bad_pattern)/good_pattern >= 0.25:
                self.test_pattern = False
            elif bad_pattern > 0:
                self.warning("Some WRITE packets were not capture properly")

        if len(xids) == 0:
            return 0

        # Flag showing if this DS is the same as the MDS
        dsismds = (ipaddr == self.server_ipaddr and port == self.port)

        # Find all I/O replies for MDS or current DS
        while True:
            # Find I/O reply
            pkt = self.pktt.match("NFS.resop == %d" % io_op, maxindex=maxindex)
            if not pkt:
                break
            xid = pkt.rpc.xid
            if xid in xids:
                xids.remove(xid)
                nfsop = pkt.NFSop

                self.test_counts.append(nfsop.count)
                xid_counts.pop(xid, None)
                if iomode == LAYOUTIOMODE4_READ:
                    offset = offsets[xid]

                    # Get real file offset
                    file_offset = self.get_abs_offset(offset, ds_index)

                    data = self.data_pattern(file_offset, len(nfsop.data), pattern=pattern)
                    if data != nfsop.data:
                        bad_pattern += 1
                    else:
                        good_pattern += 1
                else:
                    if pkt.nfs.status == NFS4_OK:
                        if not dsismds:
                            self.mdsd_lcommit = True
                        if nfsop.committed < FILE_SYNC4:
                            # Need layout commit if reply is not FILE_SYNC4
                            self.need_lcommit = True
                        if nfsop.committed == UNSTABLE4:
                            # Need commit if reply is UNSTABLE4
                            self.need_commit = True
                        if self.writeverf is None:
                            self.writeverf = nfsop.verifier
                        if self.writeverf != nfsop.verifier:
                            self.test_verf = False
                    else:
                        # Server returned error for this I/O operation
                        errstr = nfsstat4.get(pkt.nfs.status)
                        if self.error_hash.get(errstr) is None:
                            self.error_hash[errstr] = 1
                        else:
                            self.error_hash[errstr] += 1

                if len(xids) == 0:
                    break
            else:
                # Call was not found for this reply
                self.test_niomiss += 1
        # Add the number of calls with no replies
        self.test_niomiss += len(xids)
        nops = good_pattern + bad_pattern + self.test_niomiss

        # Append I/O call counts for those replies which were not found
        for count in xid_counts.values():
            self.test_counts.append(count)

        if iomode == LAYOUTIOMODE4_READ:
            self.dprint('DBG7', "READ bad/good pattern %d/%d" % (bad_pattern, good_pattern))
            if good_pattern == 0 or float(bad_pattern)/good_pattern >= 0.25:
                self.test_pattern = False
            elif bad_pattern > 0:
                self.warning("Some READ packets were not capture properly")

        if len(xids) > 0:
            self.warning("Could not find all replies to %s" % ('READ' if iomode == LAYOUTIOMODE4_READ else 'WRITE'))

        return nops

    def verify_commit(self, ipaddr, port, filehandle, init=False):
        """Verify commits are properly sent to the server specified by the
           given ipaddr and port.

           ipaddr:
               Destination IP address of MDS or DS
           port:
               Destination port number of MDS or DS
           filehandle:
               Find commits for this file handle
           init:
               Initialized test variables [default: False]

           Return the number of commits sent to the server.
        """
        dst = self.pktt.ip_tcp_dst_expr(ipaddr, port)
        fh = "NFS.fh == '%s'" % self.pktt.escape(filehandle)
        save_index = self.pktt.get_index()
        xids = []
        if init:
            self.test_commit_full = True
            self.test_no_commit   = False
            self.test_commit_verf = True

        while True:
            # Find COMMIT request for current DS
            pkt = self.pktt.match(dst + " and " + fh + " and NFS.argop == %d" % OP_COMMIT)
            if not pkt:
                break
            xids.append(pkt.rpc.xid)
            nfscommit = pkt.NFSop
            if nfscommit.offset != 0 or nfscommit.count != 0:
                self.test_commit_full = False

        ncommits = len(xids)
        if ncommits == 0:
            # No COMMIT was found
            self.test_no_commit = True
            return 0

        # Rewind trace file to saved packet index
        self.pktt.rewind(save_index)
        while True:
            # Find COMMIT reply for current DS
            pkt = self.pktt.match("NFS.resop == %d" % OP_COMMIT)
            if not pkt:
                break
            if pkt.rpc.xid in xids:
                nfscommit = pkt.NFSop
                if self.writeverf != nfscommit.verifier:
                    self.test_commit_verf = False

        return ncommits

    def verify_layoutcommit(self, filehandle, filesize):
        """Verify layoutcommit is properly sent to the server specified by
           the ipaddr for self.server and port given by self.port.
           Verify a GETATTR asking for file size is sent within the same
           compound as the LAYOUTCOMMIT.
           Verify GETATTR returns correct size for the file.

           filehandle:
               Find layoutcommit for this file handle
           filesize:
               Expected size of file
        """
        dst = self.pktt.ip_tcp_dst_expr(self.server_ipaddr, self.port)
        fh = "NFS.fh == '%s'" % self.pktt.escape(filehandle)

        # Find LAYOUTCOMMIT request
        pkt = self.pktt.match(dst + " and " + fh + " and NFS.argop == %d" % OP_LAYOUTCOMMIT)
        if self.layout['commit_mds']:
            self.test(not pkt, "LAYOUTCOMMIT should not be sent to MDS when NFL4_UFLG_COMMIT_THRU_MDS is set")
        else:
            if self.need_lcommit:
                if self.mdsd_lcommit:
                    self.test(pkt, "LAYOUTCOMMIT should be sent to MDS when NFL4_UFLG_COMMIT_THRU_MDS is not set")
                else:
                    self.test(not pkt, "LAYOUTCOMMIT should not be sent to MDS when DS == MDS")
            else:
                self.test(not pkt, "LAYOUTCOMMIT should not be sent to MDS (FILE_SYNC4)")
            if not pkt:
                return

            xid = pkt.rpc.xid
            layoutcommit = pkt.NFSop
            range_expr = layoutcommit.offset == 0 and layoutcommit.length in (filesize, NFS4_UINT64_MAX)
            self.test(range_expr, "LAYOUTCOMMIT should be sent to MDS with correct file range")
            self.test(layoutcommit.stateid == self.layout['stateid'], "LAYOUTCOMMIT should use the layout stateid")
            self.test(layoutcommit.last_write_offset.newoffset, "LAYOUTCOMMIT new offset should be set")
            self.test(layoutcommit.last_write_offset.offset == (filesize - 1),
                      "LAYOUTCOMMIT last write offset (%d) should be one less than the file size (%d)" % (layoutcommit.last_write_offset.offset, filesize))
            self.test(layoutcommit.layoutupdate.type == LAYOUT4_NFSV4_1_FILES, "LAYOUTCOMMIT layout type should be LAYOUT4_NFSV4_1_FILES")
            self.test(len(layoutcommit.layoutupdate.body) == 0, "LAYOUTCOMMIT layout update field should be empty for LAYOUT4_NFSV4_1_FILES")

            # Verify a GETATTR asking for file size is sent with LAYOUTCOMMIT
            idx = pkt.NFSidx
            getattr_arg = pkt.nfs.array[idx+1]
            self.test(getattr_arg.request & (1 << FATTR4_SIZE), "GETATTR asking for file size is sent within LAYOUTCOMMIT compound")

            # Find LAYOUTCOMMIT reply
            pkt = self.pktt.match("RPC.xid == %d and NFS.resop == %d" % (xid, OP_LAYOUTCOMMIT))
            layoutcommit = pkt.NFSop
            if layoutcommit.newsize.sizechanged:
                self.test(True, "LAYOUTCOMMIT reply file size changed should be set")
                ns_size = layoutcommit.newsize.size
                if ns_size == filesize:
                    self.test(True, "LAYOUTCOMMIT reply file size should be correct (%d)" % ns_size)
                else:
                    self.warning("LAYOUTCOMMIT reply file size is not correct (%d)" % ns_size)
            else:
                self.test(True, "LAYOUTCOMMIT reply file size changed is not set (ERRATA)")

            # Verify GETATTR returns correct file size
            idx = pkt.NFSidx
            getattr_res = pkt.nfs.array[idx+1]
            self.test(getattr_res.attributes[FATTR4_SIZE] == filesize, "GETATTR should return correct file size within LAYOUTCOMMIT compound")
        return

    def get_stateid(self, filename, **kwargs):
        """Search the packet trace for the file name given to get the OPEN
           so all related state ids can be searched. A couple of object
           attributes are defined, one is the correct state id that should
           be used by I/O operations. The second is a dictionary table
           which maps the state id to a string identifying if the state
           id is an open, lock or delegation state id.

           ipaddr:
               Destination IP address [default: self.server_ipaddr]
           port:
               Destination port [default: self.port]
           noreset:
               Do not reset the state id map [default: False]
        """
        noreset = kwargs.pop("noreset", False)
        if not noreset:
            self.stid_map = {}
        self.lock_stateid = None
        (self.filehandle, self.open_stateid, self.deleg_stateid) = self.find_open(filename=filename, **kwargs)
        if self.open_stateid:
            self.stid_map[self.open_stateid] = "OPEN stateid"
        if self.deleg_stateid:
            # Delegation stateid should be used for I/O
            self.stateid = self.deleg_stateid
            self.stid_map[self.deleg_stateid] = "DELEG stateid"
        else:
            # Look for a lock stateid
            save_index = self.pktt.get_index()
            argl = ("ipaddr", "port")
            args = dict((k, kwargs[k]) for k in kwargs if k in argl)
            args["match"] = "NFS.fh == '%s'" % self.pktt.escape(self.filehandle)
            (pktcall, pktreply) = self.find_nfs_op(OP_LOCK, **args)
            if pktreply:
                self.lock_stateid = pktreply.NFSop.stateid.other
                self.stid_map[self.lock_stateid] = "LOCK stateid"
                self.stateid = self.lock_stateid
            else:
                # Open stateid should be used for I/O
                self.stateid = self.open_stateid
            self.pktt.rewind(save_index)
        return self.stateid

    def get_clientid(self, **kwargs):
        """Return the client id for the given IP address and port number.

           ipaddr:
               Destination IP address [default: self.server_ipaddr]
           port:
               Destination port [default: self.port]
        """
        self.clientid = None
        # Find the EXCHANGE_ID packets
        self.find_nfs_op(OP_EXCHANGE_ID, **kwargs)
        if self.pktreply:
            self.clientid = self.pktreply.NFSop.clientid
            return self.clientid

    def get_sessionid(self, **kwargs):
        """Return the session id for the given IP address and port number.

           clientid:
               Search the CREATE_SESSION tied to this client id [default: None]
           ipaddr:
               Destination IP address [default: self.server_ipaddr]
           port:
               Destination port [default: self.port]
        """
        self.sessionid = None
        clientid = kwargs.pop('clientid', None)
        if clientid is not None:
            # Get the session id tied to the client id from the cache
            self.sessionid = self.sessionid_map.get(clientid)
            kwargs["match"] = "NFS.clientid == %d" % clientid
        # Find the CREATE_SESSION packets for the exchange id if given
        self.find_nfs_op(OP_CREATE_SESSION, **kwargs)
        if self.pktreply:
            # Save the session id from the reply
            self.sessionid = self.pktreply.NFSop.sessionid
            self.sessionid_map[clientid] = self.sessionid
        return self.sessionid

    def get_rootfh(self, **kwargs):
        """Return the root file handle from PUTROOTFH

           sessionid:
               Search the PUTROOTFH tied to this session id [default: None]
           ipaddr:
               Destination IP address [default: self.server_ipaddr]
           port:
               Destination port [default: self.port]
        """
        self.rootfh = None
        sessionid = kwargs.pop('sessionid', None)
        if sessionid is not None:
            fh = self.rootfh_map.get(sessionid)
            if fh is not None:
                # Return root fh found in the cache
                self.rootfh = fh
                return fh
            kwargs["match"] = "str(NFS.sessionid) == '%s'" % sessionid
        # Find the PUTROOTFH packets for the session id if given
        self.find_nfs_op(OP_PUTROOTFH, **kwargs)
        if self.pktreply:
            # Get the GETFH object from the packet
            getfh = self.getop(self.pktreply, OP_GETFH)
            if getfh:
                self.rootfh = getfh.fh
                return getfh.fh

    def get_pathfh(self, path, dirfh=None):
        """Return the file handle for the given path by searching the packet
           trace for every component in the path.
           The file handle for each component is used to search for the file
           handle in the next component.

           path:
               File system path
           dirfh:
               Directory file handle to start with [default: None]
        """
        self.pktcall  = None
        self.pktreply = None
        # Break path into its directory components
        path_list = split_path(path)
        while len(path_list):
            # Get next path component
            name = path_list.pop(0)
            if dirfh is None:
                dirmatch = ""
            else:
                dirmatch = "crc32(nfs.fh) == %d and " % crc32(dirfh)
            # Match any operation with a name attribute,
            # e.g., LOOKUP, CREATE, etc.
            mstr = "%snfs.name == '%s'" % (dirmatch, name)
            while self.pktt.match(mstr, rewind=False, reply=True):
                pkt = self.pktt.pkt
                if pkt.rpc.type == 0:
                    # Save packet call
                    self.pktcall = pkt
                else:
                    # Save packet reply
                    self.pktreply = pkt
                    if hasattr(pkt.nfs, "status") and pkt.nfs.status == 0:
                        # Get GETFH from the packet reply where name was matched
                        getfh = self.getop(pkt, OP_GETFH)
                        if getfh:
                            # Set file handle for next iteration
                            dirfh = getfh.fh
                            break
            if self.pktt.pkt is None:
                # The name was not matched, so return None
                return
        return dirfh

    def stid_str(self, stateid):
        """Display the state id in CRC16 format"""
        stid = self.format("{0:crc16}", stateid)
        return self.stid_map.get(stateid, stid)

    def get_freebytes(self, dir=None):
        """Get the number of bytes available in the given directory.
           It takes into account the effective user running the test.
           The root user is allowed to use all the available disk space
           on the device, on the other hand a regular user is allowed a
           little bit less.
        """
        if dir is None:
            dir = self.mtdir
        statvfs = os.statvfs(dir)
        if os.getuid() == 0:
            # Use free blocks if root user
            return statvfs.f_bsize * (statvfs.f_bfree-1)
        else:
            # Use free blocks available for a non-root user
            return statvfs.f_bsize * (statvfs.f_bavail-1)

    @staticmethod
    def iomode_str(iomode):
        """Return a string representation of iomode.
           This could be run as an instance or class method.
        """
        if layoutiomode4.get(iomode):
            return layoutiomode4[iomode]
        else:
            return str(iomode)

    @staticmethod
    def bitmap_str(bitmap, count, bmap, blist):
        """Return the string representation of bitmap.

           bitmap:
               Bitmap to convert
           count:
               Number of occurrences of bitmap
           bmap:
               Dictionary mapping the bits to strings
           blist:
               List of all possible bit combinations
        """
        # Get number of instances of bitmap
        cnt = 0
        for item in blist:
            if bitmap & item == bitmap:
                cnt += 1
        plist = []
        bit = max(bmap.keys())

        # Convert bitmap to a string
        while bit > 0:
            if bitmap & bit:
                plist.append(bmap[bit])
            bit = bit >> 1
        if cnt == count:
            return " & ".join(plist)
        return



#===============================================================================
# Copyright 2012 NetApp, Inc. All Rights Reserved,
# contribution by Jorge Mora <mora@netapp.com>
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#===============================================================================
"""
Pkt module

Provides the object for a packet and the string representation of the packet.
This object has an attribute for each of the layers in the packet so each layer
can be accessed directly instead of going through each layer. To access the nfs
layer object you can use 'x.nfs' instead of using 'x.ethernet.ip.tcp.rpc.nfs'
which would be very cumbersome to use. Also, since NFS can be used with either
TCP or UDP it would be harder to access the nfs object independently of
the protocol.

Packet object attributes:
    Pkt(
        record   = Record information (frame number, etc.)
        ethernet = ETHERNET II (RFC 894) object
        ip       = IPv4 object
        tcp      = TCP object
        rpc      = RPC object
        nfs      = NFS object
    )
"""
import nfstest_config as c
from baseobj import BaseObj

# Module constants
__author__    = "Jorge Mora (%s)" % c.NFSTEST_AUTHOR_EMAIL
__copyright__ = "Copyright (C) 2012 NetApp, Inc."
__license__   = "GPL v2"
__version__   = "1.1"

# The order in which to display all layers in the packet
PKT_layers = [
    'record',
    'ethernet',
    'ip', 'arp', 'rarp',
    'tcp', 'udp',
    'rpc', 'ntp', 'dns', 'krb',
    'gssd', 'nfs', 'mount', 'portmap', 'nlm', 'gssc',
]
# Required layers for debug_repr(1)
_PKT_rlayers = ['record', 'ip']
# Do not display these layers for debug_repr(1)
_PKT_nlayers = ['gssd', 'gssc']
# Packet layers to display as debug_repr(2) for debug_repr(1) if last layer
_PKT_mlayers = ['record', 'ethernet', 'ip']
_maxlen = len(max(PKT_layers, key=len))

class Pkt(BaseObj):
    """Packet object

       Usage:
           from packet.pkt import Pkt

           x = Pkt()

           # Check if this is an NFS packet
           if x == 'nfs':
               print x.nfs
    """
    _attrlist = tuple(PKT_layers)

    # Do not use BaseObj constructor to have a little bit of
    # performance improvement
    def __init__(self): pass

    def __eq__(self, other):
        """Comparison method used to determine if object has a given layer"""
        if type(other) is str:
            return getattr(self, other.lower(), None) is not None
        return False

    def __ne__(self, other):
        """Comparison method used to determine if object does not have a given layer"""
        return not self.__eq__(other)

    def __str__(self):
        """String representation of object

           The representation depends on the verbose level set by debug_repr().
           If set to 0 the generic object representation is returned.
           If set to 1 the representation of is condensed into a single line.
           It contains, the frame number, IP source and destination and/or the
           last layer:
               '1 0.386615 192.168.0.62 -> 192.168.0.17 TCP 2049 -> 708, seq: 3395733180, ack: 3294169773, ACK,SYN'
               '5 0.530957 00:0c:29:54:09:ef -> ff:ff:ff:ff:ff:ff, type: 0x806'
               '19 0.434370 192.168.0.17 -> 192.168.0.62 NFS v4 COMPOUND4 call  SEQUENCE;PUTFH;GETATTR'

           If set to 2 the representation of the object is a line for each layer:
               'Pkt(
                    RECORD:   frame 19 @ 0.434370 secs, 238 bytes on wire, 238 bytes captured
                    ETHERNET: 00:0c:29:54:09:ef -> e4:ce:8f:58:9f:f4, type: 0x800(IPv4)
                    IP:       192.168.0.17 -> 192.168.0.62, protocol: 6(TCP), len: 224
                    TCP:      src port 708 -> dst port 2049, seq: 3294170673, ack: 3395734137, len: 172, flags: ACK,PSH
                    RPC:      CALL(0), program: 100003, version: 4, procedure: 1, xid: 0x1437d3d5
                    NFS:      COMPOUND4args(tag='', minorversion=1, argarray=[nfs_argop4(argop=OP_SEQUENCE, ...), ...])
                )'
        """
        rdebug = self.debug_repr()
        if rdebug > 0:
            out = "Pkt(\n" if rdebug == 2 else ''
            klist = []
            for key in PKT_layers:
                if getattr(self, key, None) is not None and (rdebug > 1 or key not in _PKT_nlayers):
                    klist.append(key)
            index = 0
            lastkey = len(klist) - 1
            for key in klist:
                value = getattr(self, key, None)
                if value is not None and (rdebug > 1 or index == lastkey or key in _PKT_rlayers or (self != "ip" and key == "ethernet")):
                    if rdebug == 1:
                        if index == lastkey and key in _PKT_mlayers:
                            # Display level 2 if last layer is in _PKT_mlayers
                            self.debug_repr(2)
                            out += str(value)
                            self.debug_repr(1)
                        else:
                            out += str(value)
                    else:
                        sps = " " * (_maxlen - len(key))
                        out += "    %s:%s %s\n" % (key.upper(), sps, str(value))
                        if index == lastkey and hasattr(value, "data") and value.data is not None and key != "nfs":
                            sps = " " * (_maxlen - 4)
                            out += "    DATA:%s 0x%s\n" % (sps, value.data.encode("hex"))
                index += 1
            out += ")\n" if rdebug == 2 else ""
        else:
            out = BaseObj.__str__(self)
        return out



#===============================================================================
# Copyright 2012 NetApp, Inc. All Rights Reserved,
# contribution by Jorge Mora <mora@netapp.com>
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#===============================================================================
"""
IPv4 module

Decode IP version 4 layer.
"""
import struct
import nfstest_config as c
from baseobj import BaseObj
from packet.utils import ShortHex
from packet.transport.tcp import TCP
from packet.transport.udp import UDP

# Module constants
__author__    = "Jorge Mora (%s)" % c.NFSTEST_AUTHOR_EMAIL
__copyright__ = "Copyright (C) 2012 NetApp, Inc."
__license__   = "GPL v2"
__version__   = "1.1"

# Name of different protocols
_IP_map = {1:'ICMP(1)', 2:'IGMP(2)', 6:'TCP(6)', 17:'UDP(17)'}

class Flags(BaseObj):
    """Flags object"""
    # Class attributes
    _attrlist = ("DF", "MF")

    def __init__(self, data):
        """Constructor which takes a single byte as input"""
        self.DF = ((data >> 14) & 0x01) # Don't Fragment
        self.MF = ((data >> 13) & 0x01) # More Fragments

class IPv4(BaseObj):
    """IPv4 object

       Usage:
           from packet.internet.ipv4 import IPv4

           x = IPv4(pktt)

       Object definition:

       IPv4(
           version         = int,
           IHL             = int, # Internet Header Length (in 32bit words)
           header_size     = int, # IHL in actual bytes
           DSCP            = int, # Differentiated Services Code Point
           ECN             = int, # Explicit Congestion Notification
           total_size      = int, # Total length
           id              = int, # Identification
           flags = Flags(         # Flags:
               DF = int,          #   Don't Fragment
               MF = int,          #   More Fragments
           )
           fragment_offset = int, # Fragment offset (in 8-byte blocks)
           TTL             = int, # Time to Live
           protocol        = int, # Protocol of next layer (RFC790)
           checksum        = int, # Header checksum
           src             = "%d.%d.%d.%d", # source IP address
           dst             = "%d.%d.%d.%d", # destination IP address
           options = string, # IP options if available
           data = string,    # Raw data of payload if protocol
                             # is not supported
       )
    """
    # Class attributes
    _attrlist = ("version", "IHL", "header_size", "DSCP", "ECN", "total_size",
                 "id", "flags", "fragment_offset", "TTL", "protocol",
                 "checksum", "src", "dst", "options", "data")

    def __init__(self, pktt):
        """Constructor

           Initialize object's private data.

           pktt:
               Packet trace object (packet.pktt.Pktt) so this layer has
               access to the parent layers.
        """
        # Decode IP header
        unpack = pktt.unpack
        ulist = unpack.unpack(20, "!BBHHHBBH4B4B")
        self.version         = (ulist[0] >> 4)
        self.IHL             = (ulist[0] & 0x0F)
        self.header_size     = 4*self.IHL
        self.DSCP            = (ulist[1] >> 2)
        self.ECN             = (ulist[1] & 0x03)
        self.total_size      = ulist[2]
        self.id              = ShortHex(ulist[3])
        self.flags           = Flags(ulist[4])
        self.fragment_offset = (ulist[4] & 0x1FFF)
        self.TTL             = ulist[5]
        self.protocol        = ulist[6]
        self.checksum        = ShortHex(ulist[7])
        self.src             = "%d.%d.%d.%d" % ulist[8:12]
        self.dst             = "%d.%d.%d.%d" % ulist[12:]

        pktt.pkt.ip = self

        if self.header_size > 20:
            # Save IP options
            osize = self.header_size - 20
            self.options = unpack.read(osize)

        if self.protocol == 6:
            # Decode TCP
            TCP(pktt)
        elif self.protocol == 17:
            # Decode UDP
            UDP(pktt)
        else:
            self.data = unpack.getbytes()

    def __str__(self):
        """String representation of object

           The representation depends on the verbose level set by debug_repr().
           If set to 0 the generic object representation is returned.
           If set to 1 the representation of the object is condensed:
               '192.168.0.20 -> 192.168.0.61 '

           If set to 2 the representation of the object also includes the
           protocol and length of payload:
               '192.168.0.20 -> 192.168.0.61, protocol: 17(UDP), len: 84'
        """
        rdebug = self.debug_repr()
        if rdebug == 1:
            out = "%s -> %s " % (self.src, self.dst)
        elif rdebug == 2:
            proto = _IP_map.get(self.protocol, self.protocol)
            out = "%s -> %s, protocol: %s, len: %d" % (self.src, self.dst, proto, self.total_size)
        else:
            out = BaseObj.__str__(self)
        return out



#===============================================================================
# Copyright 2014 NetApp, Inc. All Rights Reserved,
# contribution by Jorge Mora <mora@netapp.com>
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#===============================================================================
# Generated by process_xdr.py from nlm4.x on Sun Oct 04 08:09:28 2015
"""
NLMv4 constants module
"""
import nfstest_config as c

# Module constants
__author__    = "Jorge Mora (%s)" % c.NFSTEST_AUTHOR_EMAIL
__copyright__ = "Copyright (C) 2014 NetApp, Inc."
__license__   = "GPL v2"
__version__   = "4.0"

# Enum nfs_bool
FALSE = 0
TRUE  = 1

nfs_bool = {
    0 : "FALSE",
    1 : "TRUE",
}

# Sizes
LM_MAXSTRLEN = 1024
MAXNAMELEN   = 1025  # LM_MAXSTRLEN + 1
MAXNETOBJ_SZ = 1024

# Enum nlm4_stats
NLM4_GRANTED             = 0
NLM4_DENIED              = 1
NLM4_DENIED_NOLOCKS      = 2
NLM4_BLOCKED             = 3
NLM4_DENIED_GRACE_PERIOD = 4
NLM4_DEADLCK             = 5
NLM4_ROFS                = 6
NLM4_STALE_FH            = 7
NLM4_FBIG                = 8
NLM4_FAILED              = 9

nlm4_stats = {
    0 : "NLM4_GRANTED",
    1 : "NLM4_DENIED",
    2 : "NLM4_DENIED_NOLOCKS",
    3 : "NLM4_BLOCKED",
    4 : "NLM4_DENIED_GRACE_PERIOD",
    5 : "NLM4_DEADLCK",
    6 : "NLM4_ROFS",
    7 : "NLM4_STALE_FH",
    8 : "NLM4_FBIG",
    9 : "NLM4_FAILED",
}

# Enum fsh4_mode
fsm_DN  = 0
fsm_DR  = 1
fsm_DW  = 2
fsm_DRW = 3

fsh4_mode = {
    0 : "fsm_DN",
    1 : "fsm_DR",
    2 : "fsm_DW",
    3 : "fsm_DRW",
}

# Enum fsh4_access
fsa_NONE = 0
fsa_R    = 1
fsa_W    = 2
fsa_RW   = 3

fsh4_access = {
    0 : "fsa_NONE",
    1 : "fsa_R",
    2 : "fsa_W",
    3 : "fsa_RW",
}

# Enum nlm_proc4
NLMPROC4_NULL        = 0
NLMPROC4_TEST        = 1
NLMPROC4_LOCK        = 2
NLMPROC4_CANCEL      = 3
NLMPROC4_UNLOCK      = 4
NLMPROC4_GRANTED     = 5
NLMPROC4_TEST_MSG    = 6
NLMPROC4_LOCK_MSG    = 7
NLMPROC4_CANCEL_MSG  = 8
NLMPROC4_UNLOCK_MSG  = 9
NLMPROC4_GRANTED_MSG = 10
NLMPROC4_TEST_RES    = 11
NLMPROC4_LOCK_RES    = 12
NLMPROC4_CANCEL_RES  = 13
NLMPROC4_UNLOCK_RES  = 14
NLMPROC4_GRANTED_RES = 15
NLMPROC4_SHARE       = 20
NLMPROC4_UNSHARE     = 21
NLMPROC4_NM_LOCK     = 22
NLMPROC4_FREE_ALL    = 23

nlm_proc4 = {
     0 : "NLMPROC4_NULL",
     1 : "NLMPROC4_TEST",
     2 : "NLMPROC4_LOCK",
     3 : "NLMPROC4_CANCEL",
     4 : "NLMPROC4_UNLOCK",
     5 : "NLMPROC4_GRANTED",
     6 : "NLMPROC4_TEST_MSG",
     7 : "NLMPROC4_LOCK_MSG",
     8 : "NLMPROC4_CANCEL_MSG",
     9 : "NLMPROC4_UNLOCK_MSG",
    10 : "NLMPROC4_GRANTED_MSG",
    11 : "NLMPROC4_TEST_RES",
    12 : "NLMPROC4_LOCK_RES",
    13 : "NLMPROC4_CANCEL_RES",
    14 : "NLMPROC4_UNLOCK_RES",
    15 : "NLMPROC4_GRANTED_RES",
    20 : "NLMPROC4_SHARE",
    21 : "NLMPROC4_UNSHARE",
    22 : "NLMPROC4_NM_LOCK",
    23 : "NLMPROC4_FREE_ALL",
}



#===============================================================================
# Copyright 2013 NetApp, Inc. All Rights Reserved,
# contribution by Jorge Mora <mora@netapp.com>
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#===============================================================================
"""
GSS constants module

Provide constant values and mapping dictionaries for the GSS layer.

RFC 2203 RPCSEC_GSS Protocol Specification
RFC 1964 The Kerberos Version 5 GSS-API Mechanism
"""
import nfstest_config as c

# Module constants
__author__    = "Jorge Mora (%s)" % c.NFSTEST_AUTHOR_EMAIL
__copyright__ = "Copyright (C) 2013 NetApp, Inc."
__license__   = "GPL v2"
__version__   = "1.2"

# Enum rpc_gss_service_t
rpc_gss_svc_none         = 1
rpc_gss_svc_integrity    = 2
rpc_gss_svc_privacy      = 3
rpc_gss_svc_channel_prot = 4
rpc_gss_service = {
    1: 'rpc_gss_svc_none',
    2: 'rpc_gss_svc_integrity',
    3: 'rpc_gss_svc_privacy',
    4: 'rpc_gss_svc_channel_prot',
}

# Enum rpc_gss_proc_t
RPCSEC_GSS_DATA          = 0
RPCSEC_GSS_INIT          = 1
RPCSEC_GSS_CONTINUE_INIT = 2
RPCSEC_GSS_DESTROY       = 3
RPCSEC_GSS_BIND_CHANNEL  = 4
rpc_gss_proc = {
    0: 'RPCSEC_GSS_DATA',
    1: 'RPCSEC_GSS_INIT',
    2: 'RPCSEC_GSS_CONTINUE_INIT',
    3: 'RPCSEC_GSS_DESTROY',
    4: 'RPCSEC_GSS_BIND_CHANNEL',
}

# Enum rgss2_bind_chan_status
RGSS2_BIND_CHAN_OK           = 0
RGSS2_BIND_CHAN_PREF_NOTSUPP = 1
RGSS2_BIND_CHAN_HASH_NOTSUPP = 2
gss_bind_chan_stat = {
    0: 'RGSS2_BIND_CHAN_OK',
    1: 'RGSS2_BIND_CHAN_PREF_NOTSUPP',
    2: 'RGSS2_BIND_CHAN_HASH_NOTSUPP',
}

RPCSEC_GSS_VERS_1 = 1
RPCSEC_GSS_VERS_2 = 2

# Integrity algorithm indicator
DES_MAC_MD5 = 0x0000
MD2_5       = 0x0100
DES_MAC     = 0x0200

gss_sgn_alg = {
    0x0000: "DES_MAC_MD5",
    0x0100: "MD2.5",
    0x0200: "DES_MAC",
}

# Enum gss_major_status
GSS_S_COMPLETE                   = 0x00000000  # Indicates an absence of any API errors or supplementary information bits
# Supplementary Information Codes
GSS_S_CONTINUE_NEEDED            = 0x00000001  # Returned only by gss_init_sec_context() or gss_accept_sec_context().
                                               # The routine must be called again to complete its function
GSSERR_S_DUPLICATE_TOKEN         = 0x00000002  # The token was a duplicate of an earlier token
GSSERR_S_OLD_TOKEN               = 0x00000004  # The token's validity period has expired
GSSERR_S_UNSEQ_TOKEN             = 0x00000008  # A later token has already been processed
GSSERR_S_GAP_TOKEN               = 0x00000010  # An expected per-message token was not received
# Routine Errors
GSSERR_S_BAD_MECH                = 0x00010000  # An unsupported mechanism was requested
GSSERR_S_BAD_NAME                = 0x00020000  # An invalid name was supplied
GSSERR_S_BAD_NAMETYPE            = 0x00030000  # A supplied name was of an unsupported type
GSSERR_S_BAD_BINDINGS            = 0x00040000  # Incorrect channel bindings were supplied
GSSERR_S_BAD_STATUS              = 0x00050000  # An invalid status code was supplied
GSSERR_S_BAD_SIG                 = 0x00060000  # A token had an invalid MIC
GSSERR_S_BAD_MIC                 = 0x00060000  # A token had an invalid MIC
GSSERR_S_NO_CRED                 = 0x00070000  # No credentials were supplied, or the credentials were unavailable or inaccessible
GSSERR_S_NO_CONTEXT              = 0x00080000  # No context has been established
GSSERR_S_DEFECTIVE_TOKEN         = 0x00090000  # A token was invalid
GSSERR_S_DEFECTIVE_CREDENTIAL    = 0x000a0000  # A credential was invalid
GSSERR_S_CREDENTIALS_EXPIRED     = 0x000b0000  # The referenced credentials have expired
GSSERR_S_CONTEXT_EXPIRED         = 0x000c0000  # The context has expired
GSSERR_S_FAILURE                 = 0x000d0000  # Miscellaneous failure. The underlying mechanism detected an error for which no
                                               # specific GSS-API status code is defined. The mechanism-specific status code
                                               # (minor-status code) provides more details about the error.
GSSERR_S_BAD_QOP                 = 0x000e0000  # The quality-of-protection requested could not be provided
GSSERR_S_UNAUTHORIZED            = 0x000f0000  # The operation is forbidden by local security policy
GSSERR_S_UNAVAILABLE             = 0x00100000  # The operation or option is unavailable
GSSERR_S_DUPLICATE_ELEMENT       = 0x00110000  # The requested credential element already exists
GSSERR_S_NAME_NOT_MN             = 0x00120000  # The provided name was not a Mechanism Name (MN)
# Calling Errors
GSSERR_S_CALL_INACCESSIBLE_READ  = 0x01000000  # A required input parameter could not be read
GSSERR_S_CALL_INACCESSIBLE_WRITE = 0x02000000  # A required output parameter could not be written
GSSERR_S_CALL_BAD_STRUCTURE      = 0x03000000  # A parameter was malformed

gss_major_status = {
    0x00000000 : "GSS_S_COMPLETE",
    0x00000001 : "GSS_S_CONTINUE_NEEDED",
    0x00000002 : "GSSERR_S_DUPLICATE_TOKEN",
    0x00000004 : "GSSERR_S_OLD_TOKEN",
    0x00000008 : "GSSERR_S_UNSEQ_TOKEN",
    0x00000010 : "GSSERR_S_GAP_TOKEN",
    0x00010000 : "GSSERR_S_BAD_MECH",
    0x00020000 : "GSSERR_S_BAD_NAME",
    0x00030000 : "GSSERR_S_BAD_NAMETYPE",
    0x00040000 : "GSSERR_S_BAD_BINDINGS",
    0x00050000 : "GSSERR_S_BAD_STATUS",
    0x00060000 : "GSSERR_S_BAD_SIG",
    0x00060000 : "GSSERR_S_BAD_MIC",
    0x00070000 : "GSSERR_S_NO_CRED",
    0x00080000 : "GSSERR_S_NO_CONTEXT",
    0x00090000 : "GSSERR_S_DEFECTIVE_TOKEN",
    0x000a0000 : "GSSERR_S_DEFECTIVE_CREDENTIAL",
    0x000b0000 : "GSSERR_S_CREDENTIALS_EXPIRED",
    0x000c0000 : "GSSERR_S_CONTEXT_EXPIRED",
    0x000d0000 : "GSSERR_S_FAILURE",
    0x000e0000 : "GSSERR_S_BAD_QOP",
    0x000f0000 : "GSSERR_S_UNAUTHORIZED",
    0x00100000 : "GSSERR_S_UNAVAILABLE",
    0x00110000 : "GSSERR_S_DUPLICATE_ELEMENT",
    0x00120000 : "GSSERR_S_NAME_NOT_MN",
    0x01000000 : "GSSERR_S_CALL_INACCESSIBLE_READ",
    0x02000000 : "GSSERR_S_CALL_INACCESSIBLE_WRITE",
    0x03000000 : "GSSERR_S_CALL_BAD_STRUCTURE",
}

# Enum gss_minor_status
KRB5KDC_ERR_NONE                 = -1765328384  # No error
KRB5KDC_ERR_NAME_EXP             = -1765328383  # Client's entry in database has expired
KRB5KDC_ERR_SERVICE_EXP          = -1765328382  # Server's entry in database has expired
KRB5KDC_ERR_BAD_PVNO             = -1765328381  # Requested protocol version not supported
KRB5KDC_ERR_C_OLD_MAST_KVNO      = -1765328380  # Client's key is encrypted in an old master key
KRB5KDC_ERR_S_OLD_MAST_KVNO      = -1765328379  # Server's key is encrypted in an old master key
KRB5KDC_ERR_C_PRINCIPAL_UNKNOWN  = -1765328378  # Client not found in Kerberos database
KRB5KDC_ERR_S_PRINCIPAL_UNKNOWN  = -1765328377  # Server not found in Kerberos database
KRB5KDC_ERR_PRINCIPAL_NOT_UNIQUE = -1765328376  # Principal has multiple entries in Kerberos database
KRB5KDC_ERR_NULL_KEY             = -1765328375  # Client or server has a null key
KRB5KDC_ERR_CANNOT_POSTDATE      = -1765328374  # Ticket is ineligible for postdating
KRB5KDC_ERR_NEVER_VALID          = -1765328373  # Requested effective lifetime is negative or too short
KRB5KDC_ERR_POLICY               = -1765328372  # KDC policy rejects request
KRB5KDC_ERR_BADOPTION            = -1765328371  # KDC can't fulfill requested option
KRB5KDC_ERR_ETYPE_NOSUPP         = -1765328370  # KDC has no support for encryption type
KRB5KDC_ERR_SUMTYPE_NOSUPP       = -1765328369  # KDC has no support for checksum type
KRB5KDC_ERR_PADATA_TYPE_NOSUPP   = -1765328368  # KDC has no support for padata type
KRB5KDC_ERR_TRTYPE_NOSUPP        = -1765328367  # KDC has no support for transited type
KRB5KDC_ERR_CLIENT_REVOKED       = -1765328366  # Client's credentials have been revoked
KRB5KDC_ERR_SERVICE_REVOKED      = -1765328365  # Credentials for server have been revoked
KRB5KDC_ERR_TGT_REVOKED          = -1765328364  # TGT has been revoked
KRB5KDC_ERR_CLIENT_NOTYET        = -1765328363  # Client not yet valid, try again later
KRB5KDC_ERR_SERVICE_NOTYET       = -1765328362  # Server not yet valid, try again later
KRB5KDC_ERR_KEY_EXP              = -1765328361  # Password has expired
KRB5KDC_ERR_PREAUTH_FAILED       = -1765328360  # Preauthentication failed
KRB5KDC_ERR_PREAUTH_REQUIRED     = -1765328359  # Additional preauthentication required
KRB5KDC_ERR_SERVER_NOMATCH       = -1765328358  # Requested server and ticket don't match
KRB5KRB_AP_ERR_BAD_INTEGRITY     = -1765328353  # Decrypt integrity check failed
KRB5KRB_AP_ERR_TKT_EXPIRED       = -1765328352  # Ticket expired
KRB5KRB_AP_ERR_TKT_NYV           = -1765328351  # Ticket not yet valid
KRB5KRB_AP_ERR_REPEAT            = -1765328350  # Request is a replay
KRB5KRB_AP_ERR_NOT_US            = -1765328349  # The ticket isn't for us
KRB5KRB_AP_ERR_BADMATCH          = -1765328348  # Ticket/authenticator do not match
KRB5KRB_AP_ERR_SKEW              = -1765328347  # Clock skew too great
KRB5KRB_AP_ERR_BADADDR           = -1765328346  # Incorrect net address
KRB5KRB_AP_ERR_BADVERSION        = -1765328345  # Protocol version mismatch
KRB5KRB_AP_ERR_MSG_TYPE          = -1765328344  # Invalid message type
KRB5KRB_AP_ERR_MODIFIED          = -1765328343  # Message stream modified
KRB5KRB_AP_ERR_BADORDER          = -1765328342  # Message out of order
KRB5KRB_AP_ERR_ILL_CR_TKT        = -1765328341  # Illegal cross-realm ticket
KRB5KRB_AP_ERR_BADKEYVER         = -1765328340  # Key version is not available
KRB5KRB_AP_ERR_NOKEY             = -1765328339  # Service key not available
KRB5KRB_AP_ERR_MUT_FAIL          = -1765328338  # Mutual authentication failed
KRB5KRB_AP_ERR_BADDIRECTION      = -1765328337  # Incorrect message direction
KRB5KRB_AP_ERR_METHOD            = -1765328336  # Alternative authentication method required
KRB5KRB_AP_ERR_BADSEQ            = -1765328335  # Incorrect sequence number in message
KRB5KRB_AP_ERR_INAPP_CKSUM       = -1765328334  # Inappropriate type of checksum in message
KRB5KRB_ERR_GENERIC              = -1765328324  # Generic error
KRB5KRB_ERR_FIELD_TOOLONG        = -1765328323  # Field is too long for this implementation
KRB5ERR_LIBOS_BADLOCKFLAG        = -1765328255  # Invalid flag for file lock mode
KRB5ERR_LIBOS_CANTREADPWD        = -1765328254  # Cannot read password
KRB5ERR_LIBOS_BADPWDMATCH        = -1765328253  # Password mismatch
KRB5ERR_LIBOS_PWDINTR            = -1765328252  # Password read interrupted
KRB5ERR_PARSE_ILLCHAR            = -1765328251  # Illegal character in component name
KRB5ERR_PARSE_MALFORMED          = -1765328250  # Malformed representation of principal
KRB5ERR_CONFIG_CANTOPEN          = -1765328249  # Can't open/find Kerberos /etc/krb5/krb5 configuration file
KRB5ERR_CONFIG_BADFORMAT         = -1765328248  # Improper format of Kerberos /etc/krb5/krb5 configuration file
KRB5ERR_CONFIG_NOTENUFSPACE      = -1765328247  # Insufficient space to return complete information
KRB5ERR_BADMSGTYPE               = -1765328246  # Invalid message type has been specified for encoding
KRB5ERR_CC_BADNAME               = -1765328245  # Credential cache name malformed
KRB5ERR_CC_UNKNOWN_TYPE          = -1765328244  # Unknown credential cache type
KRB5ERR_CC_NOTFOUND              = -1765328243  # No matching credential has been found
KRB5ERR_CC_END                   = -1765328242  # End of credential cache reached
KRB5ERR_NO_TKT_SUPPLIED          = -1765328241  # Request did not supply a ticket
KRB5KRB_AP_ERR_WRONG_PRINC       = -1765328240  # Wrong principal in request
KRB5KRB_AP_ERR_TKT_INVALID       = -1765328239  # Ticket has invalid flag set
KRB5ERR_PRINC_NOMATCH            = -1765328238  # Requested principal and ticket don't match
KRB5ERR_KDCREP_MODIFIED          = -1765328237  # KDC reply did not match expectations
KRB5ERR_KDCREP_SKEW              = -1765328236  # Clock skew too great in KDC reply
KRB5ERR_IN_TKT_REALM_MISMATCH    = -1765328235  # Client/server realm mismatch in initial ticket request
KRB5ERR_PROG_ETYPE_NOSUPP        = -1765328234  # Program lacks support for encryption type
KRB5ERR_PROG_KEYTYPE_NOSUPP      = -1765328233  # Program lacks support for key type
KRB5ERR_WRONG_ETYPE              = -1765328232  # Requested encryption type not used in message
KRB5ERR_PROG_SUMTYPE_NOSUPP      = -1765328231  # Program lacks support for checksum type
KRB5ERR_REALM_UNKNOWN            = -1765328230  # Cannot find KDC for requested realm
KRB5ERR_SERVICE_UNKNOWN          = -1765328229  # Kerberos service unknown
KRB5ERR_KDC_UNREACH              = -1765328228  # Cannot contact any KDC for requested realm
KRB5ERR_NO_LOCALNAME             = -1765328227  # No local name found for principal name
KRB5ERR_MUTUAL_FAILED            = -1765328226  # Mutual authentication failed
KRB5ERR_RC_TYPE_EXISTS           = -1765328225  # Replay cache type is already registered
KRB5ERR_RC_MALLOC                = -1765328224  # No more memory to allocate in replay cache code
KRB5ERR_RC_TYPE_NOTFOUND         = -1765328223  # Replay cache type is unknown
KRB5ERR_RC_UNKNOWN               = -1765328222  # Generic unknown RC error
KRB5ERR_RC_REPLAY                = -1765328221  # Message is a replay
KRB5ERR_RC_IO                    = -1765328220  # Replay I/O operation failed
KRB5ERR_RC_NOIO                  = -1765328219  # Replay cache type does not support non-volatile storage
KRB5ERR_RC_PARSE                 = -1765328218  # Replay cache name parse and format error
KRB5ERR_RC_IO_EOF                = -1765328217  # End-of-file on replay cache I/O
KRB5ERR_RC_IO_MALLOC             = -1765328216  # No more memory to allocate in replay cache I/O code
KRB5ERR_RC_IO_PERM               = -1765328215  # Permission denied in replay cache code
KRB5ERR_RC_IO_IO                 = -1765328214  # I/O error in replay cache i/o code
KRB5ERR_RC_IO_UNKNOWN            = -1765328213  # Generic unknown RC/IO error
KRB5ERR_RC_IO_SPACE              = -1765328212  # Insufficient system space to store replay information
KRB5ERR_TRANS_CANTOPEN           = -1765328211  # Can't open/find realm translation file
KRB5ERR_TRANS_BADFORMAT          = -1765328210  # Improper format of realm translation file
KRB5ERR_LNAME_CANTOPEN           = -1765328209  # Can't open or find
KRB5ERR_LNAME_NOTRANS            = -1765328208  # No translation is available for requested principal
KRB5ERR_LNAME_BADFORMAT          = -1765328207  # Improper format of translation database entry
KRB5ERR_CRYPTO_INTERNAL          = -1765328206  # Cryptosystem internal error
KRB5ERR_KT_BADNAME               = -1765328205  # Key table name malformed
KRB5ERR_KT_UNKNOWN_TYPE          = -1765328204  # Unknown Key table type
KRB5ERR_KT_NOTFOUND              = -1765328203  # Key table entry not found
KRB5ERR_KT_END                   = -1765328202  # End of key table reached
KRB5ERR_KT_NOWRITE               = -1765328201  # Cannot write to specified key table
KRB5ERR_KT_IOERR                 = -1765328200  # Error writing to key table
KRB5ERR_NO_TKT_IN_RLM            = -1765328199  # Cannot find ticket for requested realm
KRB5DES_ERR_BAD_KEYPAR           = -1765328198  # DES key has bad parity
KRB5DES_ERR_WEAK_KEY             = -1765328197  # DES key is a weak key
KRB5ERR_BAD_ENCTYPE              = -1765328196  # Bad encryption type
KRB5ERR_BAD_KEYSIZE              = -1765328195  # Key size is incompatible with encryption type
KRB5ERR_BAD_MSIZE                = -1765328194  # Message size is incompatible with encryption type
KRB5ERR_CC_TYPE_EXISTS           = -1765328193  # Credentials cache type is already registered
KRB5ERR_KT_TYPE_EXISTS           = -1765328192  # Key table type is already registered
KRB5ERR_CC_IO                    = -1765328191  # Credentials cache I/O operation failed
KRB5ERR_FCC_PERM                 = -1765328190  # Credentials cache file permissions incorrect
KRB5ERR_FCC_NOFILE               = -1765328189  # No credentials cache file found
KRB5ERR_FCC_INTERNAL             = -1765328188  # Internal file credentials cache error
KRB5ERR_CC_WRITE                 = -1765328187  # Error writing to credentials cache file
KRB5ERR_CC_NOMEM                 = -1765328186  # No more memory to allocate in credentials cache code
KRB5ERR_CC_FORMAT                = -1765328185  # Bad format in credentials cache
KRB5ERR_INVALID_FLAGS            = -1765328184  # Invalid KDC option combination, which is an internal library error
KRB5ERR_NO_2ND_TKT               = -1765328183  # Request missing second ticket
KRB5ERR_NOCREDS_SUPPLIED         = -1765328182  # No credentials supplied to library routine
KRB5ERR_SENDAUTH_BADAUTHVERS     = -1765328181  # Bad sendauth version was sent
KRB5ERR_SENDAUTH_BADAPPLVERS     = -1765328180  # Bad application version was sent by sendauth
KRB5ERR_SENDAUTH_BADRESPONSE     = -1765328179  # Bad response during sendauth exchange
KRB5ERR_SENDAUTH_REJECTED        = -1765328178  # Server rejected authentication during sendauth exchange
KRB5ERR_PREAUTH_BAD_TYPE         = -1765328177  # Unsupported preauthentication type
KRB5ERR_PREAUTH_NO_KEY           = -1765328176  # Required preauthentication key not supplied
KRB5ERR_PREAUTH_FAILED           = -1765328175  # Generic preauthentication failure
KRB5ERR_RCACHE_BADVNO            = -1765328174  # Unsupported format version number for replay cache
KRB5ERR_CCACHE_BADVNO            = -1765328173  # Unsupported credentials cache format version number
KRB5ERR_KEYTAB_BADVNO            = -1765328172  # Unsupported version number for key table format
KRB5ERR_PROG_ATYPE_NOSUPP        = -1765328171  # Program lacks support for address type
KRB5ERR_RC_REQUIRED              = -1765328170  # Message replay detection requires rcache parameter
KRB5_ERR_BAD_HOSTNAME            = -1765328169  # Host name cannot be canonicalized
KRB5_ERR_HOST_REALM_UNKNOWN      = -1765328168  # Cannot determine realm for host
KRB5ERR_SNAME_UNSUPP_NAMETYPE    = -1765328167  # Conversion to service principal is undefined for name type
KRB5KRB_AP_ERR_V4_REPLY          = -1765328166  # Initial Ticket response appears to be Version 4 error
KRB5ERR_REALM_CANT_RESOLVE       = -1765328165  # Cannot resolve KDC for requested realm
KRB5ERR_TKT_NOT_FORWARDABLE      = -1765328164  # The requesting ticket cannot get forwardable tickets
KRB5ERR_FWD_BAD_PRINCIPAL        = -1765328163  # Bad principal name while trying to forward credentials
KRB5ERR_GET_IN_TKT_LOOP          = -1765328162  # Looping detected inside krb5_get_in_tkt
KRB5ERR_CONFIG_NODEFREALM        = -1765328161  # Configuration file /etc/krb5/krb5.conf does not specify default realm
KRB5ERR_SAM_UNSUPPORTED          = -1765328160  # Bad SAM flags in obtain_sam_padata
KRB5ERR_KT_NAME_TOOLONG          = -1765328159  # Keytab name too long
KRB5ERR_KT_KVNONOTFOUND          = -1765328158  # Key version number for principal in key table is incorrect
KRB5ERR_CONF_NOT_CONFIGURED      = -1765328157  # Kerberos /etc/krb5/krb5.conf configuration file not configured

gss_minor_status = {
    -1765328384 : "KRB5KDC_ERR_NONE",
    -1765328383 : "KRB5KDC_ERR_NAME_EXP",
    -1765328382 : "KRB5KDC_ERR_SERVICE_EXP",
    -1765328381 : "KRB5KDC_ERR_BAD_PVNO",
    -1765328380 : "KRB5KDC_ERR_C_OLD_MAST_KVNO",
    -1765328379 : "KRB5KDC_ERR_S_OLD_MAST_KVNO",
    -1765328378 : "KRB5KDC_ERR_C_PRINCIPAL_UNKNOWN",
    -1765328377 : "KRB5KDC_ERR_S_PRINCIPAL_UNKNOWN",
    -1765328376 : "KRB5KDC_ERR_PRINCIPAL_NOT_UNIQUE",
    -1765328375 : "KRB5KDC_ERR_NULL_KEY",
    -1765328374 : "KRB5KDC_ERR_CANNOT_POSTDATE",
    -1765328373 : "KRB5KDC_ERR_NEVER_VALID",
    -1765328372 : "KRB5KDC_ERR_POLICY",
    -1765328371 : "KRB5KDC_ERR_BADOPTION",
    -1765328370 : "KRB5KDC_ERR_ETYPE_NOSUPP",
    -1765328369 : "KRB5KDC_ERR_SUMTYPE_NOSUPP",
    -1765328368 : "KRB5KDC_ERR_PADATA_TYPE_NOSUPP",
    -1765328367 : "KRB5KDC_ERR_TRTYPE_NOSUPP",
    -1765328366 : "KRB5KDC_ERR_CLIENT_REVOKED",
    -1765328365 : "KRB5KDC_ERR_SERVICE_REVOKED",
    -1765328364 : "KRB5KDC_ERR_TGT_REVOKED",
    -1765328363 : "KRB5KDC_ERR_CLIENT_NOTYET",
    -1765328362 : "KRB5KDC_ERR_SERVICE_NOTYET",
    -1765328361 : "KRB5KDC_ERR_KEY_EXP",
    -1765328360 : "KRB5KDC_ERR_PREAUTH_FAILED",
    -1765328359 : "KRB5KDC_ERR_PREAUTH_REQUIRED",
    -1765328358 : "KRB5KDC_ERR_SERVER_NOMATCH",
    -1765328353 : "KRB5KRB_AP_ERR_BAD_INTEGRITY",
    -1765328352 : "KRB5KRB_AP_ERR_TKT_EXPIRED",
    -1765328351 : "KRB5KRB_AP_ERR_TKT_NYV",
    -1765328350 : "KRB5KRB_AP_ERR_REPEAT",
    -1765328349 : "KRB5KRB_AP_ERR_NOT_US",
    -1765328348 : "KRB5KRB_AP_ERR_BADMATCH",
    -1765328347 : "KRB5KRB_AP_ERR_SKEW",
    -1765328346 : "KRB5KRB_AP_ERR_BADADDR",
    -1765328345 : "KRB5KRB_AP_ERR_BADVERSION",
    -1765328344 : "KRB5KRB_AP_ERR_MSG_TYPE",
    -1765328343 : "KRB5KRB_AP_ERR_MODIFIED",
    -1765328342 : "KRB5KRB_AP_ERR_BADORDER",
    -1765328341 : "KRB5KRB_AP_ERR_ILL_CR_TKT",
    -1765328340 : "KRB5KRB_AP_ERR_BADKEYVER",
    -1765328339 : "KRB5KRB_AP_ERR_NOKEY",
    -1765328338 : "KRB5KRB_AP_ERR_MUT_FAIL",
    -1765328337 : "KRB5KRB_AP_ERR_BADDIRECTION",
    -1765328336 : "KRB5KRB_AP_ERR_METHOD",
    -1765328335 : "KRB5KRB_AP_ERR_BADSEQ",
    -1765328334 : "KRB5KRB_AP_ERR_INAPP_CKSUM",
    -1765328324 : "KRB5KRB_ERR_GENERIC",
    -1765328323 : "KRB5KRB_ERR_FIELD_TOOLONG",
    -1765328255 : "KRB5ERR_LIBOS_BADLOCKFLAG",
    -1765328254 : "KRB5ERR_LIBOS_CANTREADPWD",
    -1765328253 : "KRB5ERR_LIBOS_BADPWDMATCH",
    -1765328252 : "KRB5ERR_LIBOS_PWDINTR",
    -1765328251 : "KRB5ERR_PARSE_ILLCHAR",
    -1765328250 : "KRB5ERR_PARSE_MALFORMED",
    -1765328249 : "KRB5ERR_CONFIG_CANTOPEN",
    -1765328248 : "KRB5ERR_CONFIG_BADFORMAT",
    -1765328247 : "KRB5ERR_CONFIG_NOTENUFSPACE",
    -1765328246 : "KRB5ERR_BADMSGTYPE",
    -1765328245 : "KRB5ERR_CC_BADNAME",
    -1765328244 : "KRB5ERR_CC_UNKNOWN_TYPE",
    -1765328243 : "KRB5ERR_CC_NOTFOUND",
    -1765328242 : "KRB5ERR_CC_END",
    -1765328241 : "KRB5ERR_NO_TKT_SUPPLIED",
    -1765328240 : "KRB5KRB_AP_ERR_WRONG_PRINC",
    -1765328239 : "KRB5KRB_AP_ERR_TKT_INVALID",
    -1765328238 : "KRB5ERR_PRINC_NOMATCH",
    -1765328237 : "KRB5ERR_KDCREP_MODIFIED",
    -1765328236 : "KRB5ERR_KDCREP_SKEW",
    -1765328235 : "KRB5ERR_IN_TKT_REALM_MISMATCH",
    -1765328234 : "KRB5ERR_PROG_ETYPE_NOSUPP",
    -1765328233 : "KRB5ERR_PROG_KEYTYPE_NOSUPP",
    -1765328232 : "KRB5ERR_WRONG_ETYPE",
    -1765328231 : "KRB5ERR_PROG_SUMTYPE_NOSUPP",
    -1765328230 : "KRB5ERR_REALM_UNKNOWN",
    -1765328229 : "KRB5ERR_SERVICE_UNKNOWN",
    -1765328228 : "KRB5ERR_KDC_UNREACH",
    -1765328227 : "KRB5ERR_NO_LOCALNAME",
    -1765328226 : "KRB5ERR_MUTUAL_FAILED",
    -1765328225 : "KRB5ERR_RC_TYPE_EXISTS",
    -1765328224 : "KRB5ERR_RC_MALLOC",
    -1765328223 : "KRB5ERR_RC_TYPE_NOTFOUND",
    -1765328222 : "KRB5ERR_RC_UNKNOWN",
    -1765328221 : "KRB5ERR_RC_REPLAY",
    -1765328220 : "KRB5ERR_RC_IO",
    -1765328219 : "KRB5ERR_RC_NOIO",
    -1765328218 : "KRB5ERR_RC_PARSE",
    -1765328217 : "KRB5ERR_RC_IO_EOF",
    -1765328216 : "KRB5ERR_RC_IO_MALLOC",
    -1765328215 : "KRB5ERR_RC_IO_PERM",
    -1765328214 : "KRB5ERR_RC_IO_IO",
    -1765328213 : "KRB5ERR_RC_IO_UNKNOWN",
    -1765328212 : "KRB5ERR_RC_IO_SPACE",
    -1765328211 : "KRB5ERR_TRANS_CANTOPEN",
    -1765328210 : "KRB5ERR_TRANS_BADFORMAT",
    -1765328209 : "KRB5ERR_LNAME_CANTOPEN",
    -1765328208 : "KRB5ERR_LNAME_NOTRANS",
    -1765328207 : "KRB5ERR_LNAME_BADFORMAT",
    -1765328206 : "KRB5ERR_CRYPTO_INTERNAL",
    -1765328205 : "KRB5ERR_KT_BADNAME",
    -1765328204 : "KRB5ERR_KT_UNKNOWN_TYPE",
    -1765328203 : "KRB5ERR_KT_NOTFOUND",
    -1765328202 : "KRB5ERR_KT_END",
    -1765328201 : "KRB5ERR_KT_NOWRITE",
    -1765328200 : "KRB5ERR_KT_IOERR",
    -1765328199 : "KRB5ERR_NO_TKT_IN_RLM",
    -1765328198 : "KRB5DES_ERR_BAD_KEYPAR",
    -1765328197 : "KRB5DES_ERR_WEAK_KEY",
    -1765328196 : "KRB5ERR_BAD_ENCTYPE",
    -1765328195 : "KRB5ERR_BAD_KEYSIZE",
    -1765328194 : "KRB5ERR_BAD_MSIZE",
    -1765328193 : "KRB5ERR_CC_TYPE_EXISTS",
    -1765328192 : "KRB5ERR_KT_TYPE_EXISTS",
    -1765328191 : "KRB5ERR_CC_IO",
    -1765328190 : "KRB5ERR_FCC_PERM",
    -1765328189 : "KRB5ERR_FCC_NOFILE",
    -1765328188 : "KRB5ERR_FCC_INTERNAL",
    -1765328187 : "KRB5ERR_CC_WRITE",
    -1765328186 : "KRB5ERR_CC_NOMEM",
    -1765328185 : "KRB5ERR_CC_FORMAT",
    -1765328184 : "KRB5ERR_INVALID_FLAGS",
    -1765328183 : "KRB5ERR_NO_2ND_TKT",
    -1765328182 : "KRB5ERR_NOCREDS_SUPPLIED",
    -1765328181 : "KRB5ERR_SENDAUTH_BADAUTHVERS",
    -1765328180 : "KRB5ERR_SENDAUTH_BADAPPLVERS",
    -1765328179 : "KRB5ERR_SENDAUTH_BADRESPONSE",
    -1765328178 : "KRB5ERR_SENDAUTH_REJECTED",
    -1765328177 : "KRB5ERR_PREAUTH_BAD_TYPE",
    -1765328176 : "KRB5ERR_PREAUTH_NO_KEY",
    -1765328175 : "KRB5ERR_PREAUTH_FAILED",
    -1765328174 : "KRB5ERR_RCACHE_BADVNO",
    -1765328173 : "KRB5ERR_CCACHE_BADVNO",
    -1765328172 : "KRB5ERR_KEYTAB_BADVNO",
    -1765328171 : "KRB5ERR_PROG_ATYPE_NOSUPP",
    -1765328170 : "KRB5ERR_RC_REQUIRED",
    -1765328169 : "KRB5_ERR_BAD_HOSTNAME",
    -1765328168 : "KRB5_ERR_HOST_REALM_UNKNOWN",
    -1765328167 : "KRB5ERR_SNAME_UNSUPP_NAMETYPE",
    -1765328166 : "KRB5KRB_AP_ERR_V4_REPLY",
    -1765328165 : "KRB5ERR_REALM_CANT_RESOLVE",
    -1765328164 : "KRB5ERR_TKT_NOT_FORWARDABLE",
    -1765328163 : "KRB5ERR_FWD_BAD_PRINCIPAL",
    -1765328162 : "KRB5ERR_GET_IN_TKT_LOOP",
    -1765328161 : "KRB5ERR_CONFIG_NODEFREALM",
    -1765328160 : "KRB5ERR_SAM_UNSUPPORTED",
    -1765328159 : "KRB5ERR_KT_NAME_TOOLONG",
    -1765328158 : "KRB5ERR_KT_KVNONOTFOUND",
    -1765328157 : "KRB5ERR_CONF_NOT_CONFIGURED",
}



#===============================================================================
# Copyright 2012 NetApp, Inc. All Rights Reserved,
# contribution by Jorge Mora <mora@netapp.com>
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#===============================================================================
"""
ETHERNET module

Decode ethernet layer (RFC 894) Ethernet II.
"""
import nfstest_config as c
from baseobj import BaseObj
from macaddr import MacAddr
from packet.internet.ipv4 import IPv4
from packet.internet.ipv6 import IPv6
from packet.internet.arp import ARP,RARP

# Module constants
__author__    = "Jorge Mora (%s)" % c.NFSTEST_AUTHOR_EMAIL
__copyright__ = "Copyright (C) 2012 NetApp, Inc."
__license__   = "GPL v2"
__version__   = "1.1"

_ETHERNET_map = {
    0x0800: 'IPv4',
    0x86dd: 'IPv6',
    0x0806: 'ARP',
    0x8035: 'RARP',
}

class ETHERNET(BaseObj):
    """Ethernet object

       Usage:
           from packet.link.ethernet import ETHERNET

           x = ETHERNET(pktt)

       Object definition:

       ETHERNET(
           dst   = MacAddr(),  # destination MAC address
           src   = MacAddr(),  # source MAC address
           type  = int,        # payload type
           data  = string,     # raw data of payload if type is not supported
       )
    """
    # Class attributes
    _attrlist = ("dst", "src", "type", "data")

    def __init__(self, pktt):
        """Constructor

           Initialize object's private data.

           pktt:
               Packet trace object (packet.pktt.Pktt) so this layer has
               access to the parent layers.
        """
        unpack = pktt.unpack
        ulist = unpack.unpack(14, "!6s6sH")
        self.dst  = MacAddr(ulist[0].encode('hex'))
        self.src  = MacAddr(ulist[1].encode('hex'))
        self.type = ulist[2]
        pktt.pkt.ethernet = self

        if self.type == 0x0800:
            # Decode IPv4 packet
            IPv4(pktt)
        elif self.type == 0x86dd:
            # Decode IPv6 packet
            IPv6(pktt)
        elif self.type == 0x0806:
            # Decode ARP packet
            ARP(pktt)
        elif self.type == 0x8035:
            # Decode RARP packet
            RARP(pktt)
        else:
            self.data = unpack.getbytes()

    def __str__(self):
        """String representation of object

           The representation depends on the verbose level set by debug_repr().
           If set to 0 the generic object representation is returned.
           If set to 1 the representation of the object is condensed:
               '00:0c:29:54:09:ef -> 60:33:4b:29:6e:9d '

           If set to 2 the representation of the object also includes the type
           of payload:
               '00:0c:29:54:09:ef -> 60:33:4b:29:6e:9d, type: 0x800(IPv4)'
        """
        rdebug = self.debug_repr()
        if rdebug == 1:
            out = "%s -> %s " % (self.src, self.dst)
        elif rdebug == 2:
            etype = _ETHERNET_map.get(self.type, None)
            etype = hex(self.type) if etype is None else "%s(%s)" % (hex(self.type), etype)
            out = "%s -> %s, type: %s" % (self.src, self.dst, etype)
        else:
            out = BaseObj.__str__(self)
        return out



#===============================================================================
# Copyright 2014 NetApp, Inc. All Rights Reserved,
# contribution by Jorge Mora <mora@netapp.com>
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#===============================================================================
# Generated by process_xdr.py from nfs3.x on Sun Oct 04 08:09:27 2015
"""
NFSv3 constants module
"""
import nfstest_config as c

# Module constants
__author__    = "Jorge Mora (%s)" % c.NFSTEST_AUTHOR_EMAIL
__copyright__ = "Copyright (C) 2014 NetApp, Inc."
__license__   = "GPL v2"
__version__   = "3.0"

# Enum nfs_bool
FALSE = 0
TRUE  = 1

nfs_bool = {
    0 : "FALSE",
    1 : "TRUE",
}

# Sizes
NFS3_FHSIZE         = 64
NFS3_COOKIEVERFSIZE = 8
NFS3_CREATEVERFSIZE = 8
NFS3_WRITEVERFSIZE  = 8

# Enum nfsstat3
NFS3_OK             = 0
NFS3ERR_PERM        = 1
NFS3ERR_NOENT       = 2
NFS3ERR_IO          = 5
NFS3ERR_NXIO        = 6
NFS3ERR_ACCES       = 13
NFS3ERR_EXIST       = 17
NFS3ERR_XDEV        = 18
NFS3ERR_NODEV       = 19
NFS3ERR_NOTDIR      = 20
NFS3ERR_ISDIR       = 21
NFS3ERR_INVAL       = 22
NFS3ERR_FBIG        = 27
NFS3ERR_NOSPC       = 28
NFS3ERR_ROFS        = 30
NFS3ERR_MLINK       = 31
NFS3ERR_NAMETOOLONG = 63
NFS3ERR_NOTEMPTY    = 66
NFS3ERR_DQUOT       = 69
NFS3ERR_STALE       = 70
NFS3ERR_REMOTE      = 71
NFS3ERR_BADHANDLE   = 10001
NFS3ERR_NOT_SYNC    = 10002
NFS3ERR_BAD_COOKIE  = 10003
NFS3ERR_NOTSUPP     = 10004
NFS3ERR_TOOSMALL    = 10005
NFS3ERR_SERVERFAULT = 10006
NFS3ERR_BADTYPE     = 10007
NFS3ERR_JUKEBOX     = 10008

nfsstat3 = {
        0 : "NFS3_OK",
        1 : "NFS3ERR_PERM",
        2 : "NFS3ERR_NOENT",
        5 : "NFS3ERR_IO",
        6 : "NFS3ERR_NXIO",
       13 : "NFS3ERR_ACCES",
       17 : "NFS3ERR_EXIST",
       18 : "NFS3ERR_XDEV",
       19 : "NFS3ERR_NODEV",
       20 : "NFS3ERR_NOTDIR",
       21 : "NFS3ERR_ISDIR",
       22 : "NFS3ERR_INVAL",
       27 : "NFS3ERR_FBIG",
       28 : "NFS3ERR_NOSPC",
       30 : "NFS3ERR_ROFS",
       31 : "NFS3ERR_MLINK",
       63 : "NFS3ERR_NAMETOOLONG",
       66 : "NFS3ERR_NOTEMPTY",
       69 : "NFS3ERR_DQUOT",
       70 : "NFS3ERR_STALE",
       71 : "NFS3ERR_REMOTE",
    10001 : "NFS3ERR_BADHANDLE",
    10002 : "NFS3ERR_NOT_SYNC",
    10003 : "NFS3ERR_BAD_COOKIE",
    10004 : "NFS3ERR_NOTSUPP",
    10005 : "NFS3ERR_TOOSMALL",
    10006 : "NFS3ERR_SERVERFAULT",
    10007 : "NFS3ERR_BADTYPE",
    10008 : "NFS3ERR_JUKEBOX",
}

# Enum ftype3
NF3REG  = 1
NF3DIR  = 2
NF3BLK  = 3
NF3CHR  = 4
NF3LNK  = 5
NF3SOCK = 6
NF3FIFO = 7

ftype3 = {
    1 : "NF3REG",
    2 : "NF3DIR",
    3 : "NF3BLK",
    4 : "NF3CHR",
    5 : "NF3LNK",
    6 : "NF3SOCK",
    7 : "NF3FIFO",
}

# Enum time_how
DONT_CHANGE        = 0
SET_TO_SERVER_TIME = 1
SET_TO_CLIENT_TIME = 2

time_how = {
    0 : "DONT_CHANGE",
    1 : "SET_TO_SERVER_TIME",
    2 : "SET_TO_CLIENT_TIME",
}

# ACCESS3res NFSPROC3_ACCESS(ACCESS3args) = 4;
ACCESS3_READ    = 0x0001
ACCESS3_LOOKUP  = 0x0002
ACCESS3_MODIFY  = 0x0004
ACCESS3_EXTEND  = 0x0008
ACCESS3_DELETE  = 0x0010
ACCESS3_EXECUTE = 0x0020

# Enum stable_how
UNSTABLE  = 0
DATA_SYNC = 1
FILE_SYNC = 2

stable_how = {
    0 : "UNSTABLE",
    1 : "DATA_SYNC",
    2 : "FILE_SYNC",
}

# Enum createmode3
UNCHECKED = 0
GUARDED   = 1
EXCLUSIVE = 2

createmode3 = {
    0 : "UNCHECKED",
    1 : "GUARDED",
    2 : "EXCLUSIVE",
}

# FSINFO3res NFSPROC3_FSINFO(FSINFO3args) = 19;
FSF3_LINK        = 0x0001
FSF3_SYMLINK     = 0x0002
FSF3_HOMOGENEOUS = 0x0008
FSF3_CANSETTIME  = 0x0010

# Enum nfs_proc3
NFSPROC3_NULL        = 0
NFSPROC3_GETATTR     = 1
NFSPROC3_SETATTR     = 2
NFSPROC3_LOOKUP      = 3
NFSPROC3_ACCESS      = 4
NFSPROC3_READLINK    = 5
NFSPROC3_READ        = 6
NFSPROC3_WRITE       = 7
NFSPROC3_CREATE      = 8
NFSPROC3_MKDIR       = 9
NFSPROC3_SYMLINK     = 10
NFSPROC3_MKNOD       = 11
NFSPROC3_REMOVE      = 12
NFSPROC3_RMDIR       = 13
NFSPROC3_RENAME      = 14
NFSPROC3_LINK        = 15
NFSPROC3_READDIR     = 16
NFSPROC3_READDIRPLUS = 17
NFSPROC3_FSSTAT      = 18
NFSPROC3_FSINFO      = 19
NFSPROC3_PATHCONF    = 20
NFSPROC3_COMMIT      = 21

nfs_proc3 = {
     0 : "NFSPROC3_NULL",
     1 : "NFSPROC3_GETATTR",
     2 : "NFSPROC3_SETATTR",
     3 : "NFSPROC3_LOOKUP",
     4 : "NFSPROC3_ACCESS",
     5 : "NFSPROC3_READLINK",
     6 : "NFSPROC3_READ",
     7 : "NFSPROC3_WRITE",
     8 : "NFSPROC3_CREATE",
     9 : "NFSPROC3_MKDIR",
    10 : "NFSPROC3_SYMLINK",
    11 : "NFSPROC3_MKNOD",
    12 : "NFSPROC3_REMOVE",
    13 : "NFSPROC3_RMDIR",
    14 : "NFSPROC3_RENAME",
    15 : "NFSPROC3_LINK",
    16 : "NFSPROC3_READDIR",
    17 : "NFSPROC3_READDIRPLUS",
    18 : "NFSPROC3_FSSTAT",
    19 : "NFSPROC3_FSINFO",
    20 : "NFSPROC3_PATHCONF",
    21 : "NFSPROC3_COMMIT",
}
